"""This file contains custom Exceptions for pysradb
"""


class MissingQueryException(Exception):
    """Exception raised when the user did not supply any query fields.

    Attributes:
        message: string
            Error message for this Exception

    """

    def __init__(self):
        self.message = (
            "No valid query has been supplied. \n"
            "A query must be supplied to one of the following fields:\n"
            "[--query, --accession, --organism, --layout, --mbases, --publication-date,"
            " --platform, --selection, --source, --strategy, --title]"
        )
        super().__init__(self.message)


class IncorrectFieldException(Exception):
    """Exception raised when the user enters incorrect inputs for a flag."""

    pass



import os
import sqlite3
import sys
import warnings

import pandas as pd

from .utils import _extract_first_field

warnings.simplefilter(action="ignore", category=FutureWarning)


class BASEdb(object):
    def __init__(self, sqlite_file):
        """Initialize SRAdb.

        Parameters
        ----------

        sqlite_file: string
                     Path to unzipped SRAmetadb.sqlite file


        """
        self.sqlite_file = sqlite_file
        self.open()
        self.cursor = self.db.cursor()

    def open(self):
        """Open sqlite connection."""
        # Originally sqlite3.connect(self.sqlite_file)
        self.db = sqlite3.connect("file:{}?mode=ro".format(self.sqlite_file), uri=True)
        self.db.text_factory = str

    def close(self):
        """Close sqlite connection."""
        self.db.close()

    def list_tables(self):
        """List all tables in the sqlite file.

        Returns
        -------
        table_list: list
                    List of all table names
        """
        results = self.cursor.execute(
            'SELECT name FROM sqlite_master WHERE type="table";'
        ).fetchall()
        return _extract_first_field(results)

    def list_fields(self, table):
        """List all fields in a given table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        field_list: list
                    A list of field names for the table
        """
        results = self.cursor.execute("SELECT * FROM {}".format(table))
        return _extract_first_field(results.description)

    def desc_table(self, table):
        """Describe all fields in a table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        table_desc: DataFrame
                    A DataFrame with field name and its
                    schema description
        """
        results = self.cursor.execute(
            'PRAGMA table_info("{}")'.format(table)
        ).fetchall()
        columns = ["cid", "name", "dtype", "notnull", "dflt_value", "pk"]
        data = []
        for result in results:
            data.append(list([str(x) for x in result]))
        table_desc = pd.DataFrame(data, columns=columns)
        return table_desc

    def query(self, sql_query):
        """Run SQL query.

        Parameters
        ----------
        sql_query: string
                   SQL query string

        Returns
        -------
        results: DataFrame
                 Query results formatted as dataframe

        """
        results = self.cursor.execute(sql_query).fetchall()
        column_names = list([x[0] for x in self.cursor.description])
        results = [dict(list(zip(column_names, result))) for result in results]
        df = pd.DataFrame(results)
        if not results:
            # sys.stderr.write("Found no matching results for query: {}".format(sql_query))
            sys.stderr.write("Found no matching results for query.\n")
        return df

    def get_row_count(self, table):
        """Get row counts for a table.

        Parameters
        ----------
        table: string
               Table name.
               See `list_tables` for getting all table names

        Returns
        -------
        row_count: int
                   Number of rows in table
        """
        return self.cursor.execute(
            "SELECT max(rowid) FROM {}".format(table)
        ).fetchone()[0]

    def all_row_counts(self):
        """Get row counts of all tables in the db file.

        Returns
        -------
        row_counts: DataFrame
                    A dataframe with table names and corresponding
                    row count.

        """
        tables = self.list_tables()
        results = dict([(table, self.get_row_count(table)) for table in tables])
        return pd.DataFrame.from_dict(results, orient="index", columns=["count"])



"""Tests for cli.py
"""

import os
import subprocess
import sys
from shlex import quote
from shlex import split

import pytest

from pysradb import SRAdb


def run(command):
    if sys.version_info.minor >= 7:
        result = subprocess.run(split(command), capture_output=True)
    else:
        result = subprocess.run(split(command), check=True, stdout=subprocess.PIPE)
    return str(result.stdout).strip()


@pytest.fixture(scope="module")
def sradb_connection(conf_download_sradb_file):
    db_file = conf_download_sradb_file
    db = SRAdb(db_file)
    return db


def test_all_row_counts_sra(sradb_connection):
    assert sradb_connection.all_row_counts().loc["metaInfo", "count"] == 2


@pytest.mark.xfail
def test_download():
    result = run(
        "pysradb download -y --db data/SRAmetadb.sqlite --out-dir srp_downloads -p SRP063852"
    )
    assert "SRP063852" in result
    assert os.path.getsize("srp_downloads/SRP063852/SRX1254413/SRR2433794.sra")


def test_sra_metadata():
    result = run("pysradb metadata SRP098789 --db data/SRAmetadb.sqlite")
    assert "SRX2536403" in result


def test_sra_metadata():
    result = run(
        "pysradb metadata SRP098789 --db data/SRAmetadb.sqlite --detailed --expand"
    )
    assert "treatment_time" in result


def test_srp_to_srx():
    result = run("pysradb srp-to-srx SRP098789 --db data/SRAmetadb.sqlite")
    assert "SRX2536403" in result


def test_srp_assay():
    result = run("pysradb metadata SRP098789 --db data/SRAmetadb.sqlite --assay")
    assert "RNA-Seq" in result


def srr_to_srx():
    result = run(
        "pysradb srr-to-srx --db data/SRAmetadb.sqlite SRR5227288 SRR649752 --desc"
    )
    assert "3T3 cells" in result


def srx_to_srr():
    result = run(
        "pysradb srr-to-srx --db data/SRAmetadb.sqlite SRX217956 SRX2536403 --desc"
    )
    assert "3T3 cells" in result


def test_sra_metadata_detail():
    result = run(
        "pysradb metadata --db data/SRAmetadb.sqlite SRP075720 --detailed --expand"
    )
    assert "retina" in result


def test_srp_to_gse():
    result = run("pysradb srp-to-gse --db data/SRAmetadb.sqlite SRP075720")
    assert "GSE81903" in result


def test_gsm_to_srp():
    result = run("pysradb gsm-to-srp --db data/SRAmetadb.sqlite GSM2177186")
    assert "SRP075720" in result


def test_gsm_to_gse():
    result = run("pysradb gsm-to-gse --db data/SRAmetadb.sqlite GSM2177186")
    assert "GSE81903" in result


def test_gsm_to_srr():
    result = run(
        "pysradb gsm-to-srr --db data/SRAmetadb.sqlite GSM2177186 --detailed --desc --expand"
    )
    assert "GSM2177186_r1" in result


"""
def test_assay_uniq():
    result = subprocess.check_output(
        "pysradb metadata SRP000941 --db data/SRAmetadb.sqlite --assay  | "
        + " tr -s {}".format(quote("  "))
        + " | cut -f5 -d {}".format(quote(" "))
        + " | sort | uniq -c",
        shell=True,
    )
    assert "Bisulfite-Seq" in str(result)


def test_pipe_download():
    result = subprocess.check_output(
        "pysradb metadata SRP000941 --assay | "
        + " grep {}".format(quote("study\|RNA-Seq"))
        + " | head -2 | pysradb download --out-dir srp_downloads",
        shell=True,
    )
    assert os.path.getsize("srp_downloads/SRP000941/SRX007165/SRR020287.sra")
    assert "following" in str(result)
"""



"""Command line interface for pysradb
"""
import argparse
import os
import re
import sys
import warnings
from io import StringIO
from textwrap import dedent

import pandas as pd

from . import __version__
from .exceptions import IncorrectFieldException
from .exceptions import MissingQueryException
from .geoweb import GEOweb
from .search import EnaSearch
from .search import GeoSearch
from .search import SraSearch
from .sradb import SRAdb
from .sradb import download_sradb_file
from .sraweb import SRAweb
from .utils import confirm

warnings.simplefilter(action="ignore", category=FutureWarning)


class CustomFormatterArgP(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter
):
    pass


class ArgParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n" % message)
        self.print_help()
        sys.exit(2)


def _print_save_df(df, saveto=None):
    if saveto:
        if saveto.lower().endswith(".csv"):
            #  if the save file format is csv
            df.to_csv(saveto, index=False, header=True)
        else:
            df.to_csv(saveto, index=False, header=True, sep="\t")
    else:
        if df is None:
            print
        elif len(df.index):
            to_print = (
                df.replace(r"[\s]{2,}|\t", " ", regex=True)
                .to_string(
                    index=False,
                    justify="left",
                    header=False,
                    col_space=0,
                )
                .lstrip()
            )
            to_print_split = to_print.split("\n")
            to_print_split = map(lambda x: re.sub(r"\s\s+", "\t", x), to_print_split)
            to_print = ["\t".join(df.columns)]
            for line in to_print_split:
                to_print.append(line.lstrip())
            to_print = list(
                map(lambda line: line.encode("ascii", "replace").decode(), to_print)
            )
            for line in to_print:
                sys.stdout.write(line + os.linesep)
            sys.stdout.flush()


###################### metadata ##############################
def metadata(srp_id, assay, desc, detailed, expand, saveto):
    sradb = SRAweb()
    df = sradb.sra_metadata(
        srp_id,
        assay=assay,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


################################################################


################# download ##########################
def download(
    out_dir, srx, srp, geo, skip_confirmation, col="sra_url", use_ascp=False, threads=1
):
    if out_dir is None:
        out_dir = os.path.join(os.getcwd(), "pysradb_downloads")
    sradb = SRAweb()
    geoweb = GEOweb()
    # This block is triggered only if no -p or -g arguments are provided.
    # In this case, the input is taken from the pipe and assumed to be SRA, not GEO
    # TODO: at some point, we need to fix this
    if not srp and not geo:
        text = ""
        for index, line in enumerate(sys.stdin):
            line = line.strip(" \t\n\r")
            line = re.sub(r"\s*\t+\s*", "\t", line)
            if not line:
                continue
            text += "{}\n".format(line)
        df = pd.read_csv(StringIO(text), sep="\t")
        sradb.download(
            df=df,
            out_dir=out_dir,
            filter_by_srx=srx,
            skip_confirmation=True,
            use_ascp=use_ascp,
            url_col=col,
            threads=threads,
        )
    # This block is triggered for downloads using the -p argument
    if srp:
        for srp_x in srp:
            metadata = sradb.sra_metadata(srp_x, detailed=True)
            sradb.download(
                df=metadata,
                out_dir=out_dir,
                filter_by_srx=srx,
                skip_confirmation=skip_confirmation,
                use_ascp=use_ascp,
                threads=threads,
            )
    # This block is triggered for downloads using the -g argument
    if geo:
        for geo_x in geo:
            links, root_url = geoweb.get_download_links(geo_x)
            geoweb.download(links=links, root_url=root_url, gse=geo_x, out_dir=out_dir)
    sradb.close()


#########################################################


######################### search #################################
def search(saveto, db, verbosity, return_max, fields):
    if fields["run_description"]:
        verbosity = 1
    if fields["detailed"]:
        verbosity = 3
    try:
        if db == "ena":
            instance = EnaSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
            )
            instance.search()
        elif db == "geo":
            instance = GeoSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
                fields["geo_query"],
                fields["geo_dataset_type"],
                fields["geo_entry_type"],
            )
            instance.search()
        else:
            instance = SraSearch(
                verbosity,
                return_max,
                fields["query"],
                fields["accession"],
                fields["organism"],
                fields["layout"],
                fields["mbases"],
                fields["publication_date"],
                fields["platform"],
                fields["selection"],
                fields["source"],
                fields["strategy"],
                fields["title"],
            )
            instance.search()
    except (MissingQueryException, IncorrectFieldException) as e:
        print(e)
        return
    if fields["stats"]:
        instance.show_result_statistics()
    if fields["graphs"]:
        graph_types = tuple(fields["graphs"].split())
        instance.visualise_results(graph_types, False)
    _print_save_df(instance.get_df(), saveto)


def get_geo_search_info():
    print(GeoSearch.info())


####################################################################


######################### gse-to-gsm ###############################
def gse_to_gsm(gse_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gse_to_gsm(
        gse_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


####################################################################


######################## gse-to-srp ################################
def gse_to_srp(gse_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gse_to_srp(
        gse_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


######################################################################


######################### gsm-to-gse #################################
def gsm_to_gse(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_gse(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srp ################################
def gsm_to_srp(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srp(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srr ################################
def gsm_to_srr(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srr(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################ gsm-to-srs ################################
def gsm_to_srs(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srs(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


############################# gsm-to-srx ###############################
def gsm_to_srx(gsm_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.gsm_to_srx(
        gsm_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-gse ##################################
def srp_to_gse(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_gse(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srr ##################################
def srp_to_srr(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srr(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srs ##################################
def srp_to_srs(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srs(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srp-to-srx ##################################
def srp_to_srx(srp_id, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srp_to_srx(
        srp_id,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-gsm ##################################
def srr_to_gsm(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_gsm(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


########################################################################


########################### srr-to-srp ##################################
def srr_to_srp(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srp(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-srs ##################################
def srr_to_srs(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srs(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srr-to-srx ##################################
def srr_to_srx(srr_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srr_to_srx(
        srr_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srs-to-gsm ##################################
def srs_to_gsm(srs_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srs_to_gsm(
        srs_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srs-to-srx ##################################
def srs_to_srx(srs_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srs_to_srx(
        srs_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srp ##################################
def srx_to_srp(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srp(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srr ##################################
def srx_to_srr(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srr(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


########################### srx-to-srs ##################################
def srx_to_srs(srx_ids, saveto, detailed, desc, expand):
    sradb = SRAweb()
    df = sradb.srx_to_srs(
        srx_ids,
        detailed=detailed,
        sample_attribute=desc,
        expand_sample_attributes=expand,
    )
    _print_save_df(df, saveto)
    sradb.close()


#########################################################################


def parse_args(args=None):
    """Argument parser"""
    parser = ArgParser(
        description=dedent(
            """\
    pysradb: Query NGS metadata and data from NCBI Sequence Read Archive.
    version: {}.
    Citation: 10.12688/f1000research.18676.1""".format(
                __version__
            )
        ),
        formatter_class=CustomFormatterArgP,
    )

    # pysradb subcommands
    subparsers = parser.add_subparsers(title="subcommands", dest="command")

    # pysradb --version
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )

    # pysradb --citation
    parser.add_argument(
        "--citation",
        action="version",
        version=dedent(
            """
        Choudhary, Saket. "pysradb: A Python Package to Query next-Generation Sequencing Metadata and Data from NCBI Sequence Read Archive." F1000Research, vol. 8, F1000 (Faculty of 1000 Ltd), Apr. 2019, p. 532 (https://f1000research.com/articles/8-532/v1)

        @article{Choudhary2019,
        doi = {10.12688/f1000research.18676.1},
        url = {https://doi.org/10.12688/f1000research.18676.1},
        year = {2019},
        month = apr,
        publisher = {F1000 (Faculty of 1000 Ltd)},
        volume = {8},
        pages = {532},
        author = {Saket Choudhary},
        title = {pysradb: A {P}ython package to query next-generation sequencing metadata and data from {NCBI} {S}equence {R}ead {A}rchive},
        journal = {F1000Research}
        }
        """
        ),
        help="how to cite",
    )

    # pysradb metadata
    subparser = subparsers.add_parser(
        "metadata", help="Fetch metadata for SRA project (SRPnnnn)"
    )
    subparser.add_argument("--saveto", help="Save metadata dataframe to file")
    subparser.add_argument(
        "--assay", action="store_true", help="Include assay type in output"
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed", action="store_true", help="Display detailed metadata table"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id", nargs="+")
    subparser.set_defaults(func=metadata)

    # pysradb download
    subparser = subparsers.add_parser("download", help="Download SRA project (SRPnnnn)")
    subparser.add_argument("--out-dir", help="Output directory root")
    subparser.add_argument("--srx", "-x", help="Download only these SRX(s)", nargs="+")
    subparser.add_argument("--srp", "-p", help="SRP ID", nargs="+")
    subparser.add_argument("--geo", "-g", help="GEO ID", nargs="+")
    subparser.add_argument(
        "--skip-confirmation", "-y", action="store_true", help="Skip confirmation"
    )
    subparser.add_argument(
        "--use_ascp", "-a", action="store_true", help="Use aspera instead of wget"
    )
    subparser.add_argument("--col", help="Specify column to download")
    subparser.add_argument(
        "--threads", "-t", help="Number of threads", default=1, type=int
    )
    subparser.set_defaults(func=download)

    # pysradb search
    subparser = subparsers.add_parser("search", help="Search SRA/ENA for matching text")
    subparser.add_argument(
        "-o", "--saveto", help="Save search result dataframe to file"
    )
    subparser.add_argument(
        "-s",
        "--stats",
        action="store_true",
        help="Displays some useful statistics for the search results.",
    )
    subparser.add_argument(
        "-g",
        "--graphs",
        nargs="?",
        const="all",
        help=(
            "Generates graphs to illustrate the search result. "
            "By default all graphs are generated. \n"
            "Alternatively, select a subset from the options below in a space-separated string:\n"
            "daterange, organism, source, selection, platform, basecount"
        ),
    )
    subparser.add_argument(
        "-d",
        "--db",
        choices=["ena", "geo", "sra"],
        default="sra",
        help="Select the db API (sra, ena, or geo) to query, default = sra.     "
        "Note: pysradb search works slightly differently when db = geo. \n"
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-v",
        "--verbosity",
        choices=[0, 1, 2, 3],
        default=2,
        help=(
            "Level of search result details (0, 1, 2 or 3), default = 2\n"
            "0: run accession only\n"
            "1: run accession and experiment title\n"
            "2: accession numbers, titles and sequencing information\n"
            "3: records in 2 and other information such as download url, sample attributes, etc"
        ),
        type=int,
    )
    subparser.add_argument(
        "--run-description",
        action="store_true",
        help="Displays run accessions and descriptions only. Equivalent to --verbosity 1",
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Displays detailed search results. Equivalent to --verbosity 3.",
    )
    subparser.add_argument(
        "-m",
        "--max",
        default=20,
        help="Maximum number of entries to return, default = 20",
        type=int,
    )
    subparser.add_argument(
        "-q",
        "--query",
        nargs="+",
        help="Main query string. Note that if no query is supplied, at least one of the "
        "following flags must be present:",
    )
    subparser.add_argument("-A", "--accession", help="Accession number")
    subparser.add_argument(
        "-O", "--organism", nargs="+", help="Scientific name of the sample organism"
    )
    subparser.add_argument(
        "-L",
        "--layout",
        choices=["SINGLE", "PAIRED"],
        help="Library layout. Accepts either SINGLE or PAIRED",
        type=str.upper,
    )
    subparser.add_argument(
        "-M",
        "--mbases",
        help="Size of the sample rounded to the nearest megabase",
        type=int,
    )
    subparser.add_argument(
        "-D",
        "--publication-date",
        help="Publication date of the run in the format dd-mm-yyyy. If a date range is desired, "
        "enter the start date, followed by end date, separated by a colon ':'.\n "
        "Example: 01-01-2010:31-12-2010",
    )
    subparser.add_argument("-P", "--platform", nargs="+", help="Sequencing platform")
    subparser.add_argument("-E", "--selection", nargs="+", help="Library selection")
    subparser.add_argument("-C", "--source", nargs="+", help="Library source")
    subparser.add_argument(
        "-S", "--strategy", nargs="+", help="Library preparation strategy"
    )
    subparser.add_argument("-T", "--title", nargs="+", help="Experiment title")

    # The following arguments are for GEO DataSets only
    subparser.add_argument(
        "-I",
        "--geo-info",
        action="store_true",
        help="Displays information on how to query GEO DataSets via 'pysradb search --db geo ...', "
        "including accepted inputs for -G/--geo-query, -Y/--geo-dataset-type and -Z/--geo-entry-type. ",
    )
    subparser.add_argument(
        "-G",
        "--geo-query",
        nargs="+",
        help="Main query string for GEO DataSet. This flag is only used when db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-Y",
        "--geo-dataset-type",
        nargs="+",
        help="GEO DataSet Type. This flag is only used when --db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )
    subparser.add_argument(
        "-Z",
        "--geo-entry-type",
        nargs="+",
        help="GEO Entry Type. This flag is only used when --db is set to be geo."
        "Please refer to 'pysradb search --geo-info' for more details.",
    )

    subparser.set_defaults(func=search)

    # pysradb gse-to-gsm
    subparser = subparsers.add_parser("gse-to-gsm", help="Get GSM for a GSE")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                         run_accession (SRR),
                                         sample_alias (GSM),
                                         run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gse_ids", nargs="+")
    subparser.set_defaults(func=gse_to_gsm)

    # pysradb gse-to-srp
    subparser = subparsers.add_parser("gse-to-srp", help="Get SRP for a GSE")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                           run_accession (SRR),
                                           sample_accession (SRS),
                                           experiment_alias (GSM_),
                                           run_alias (GSM_r),
                                           sample_alias (GSM)]
                                           """,
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gse_ids", nargs="+")
    subparser.set_defaults(func=gse_to_gsm)

    # pysradb gsm-to-gse
    subparser = subparsers.add_parser("gsm-to-gse", help="Get GSE for a GSM")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            run_accession (SRR),
                                            sample_alias (GSM),
                                            run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_gse)

    # pysradb gsm-to-srp
    subparser = subparsers.add_parser("gsm-to-srp", help="Get SRP for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srp)

    # pysradb gsm-to-srr
    subparser = subparsers.add_parser("gsm-to-srr", help="Get SRR for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            study_accession (SRS),
                                            run_alias (GSM_r),
                                            sample_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srr)

    # pysradb gsm-to-srs
    subparser = subparsers.add_parser("gsm-to-srs", help="Get SRS for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            run_accession (SRR),
                                            study_accession (SRP),
                                            run_alias (GSM_r),
                                            experiment_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srs)

    # pysradb gsm-to-srx
    subparser = subparsers.add_parser("gsm-to-srx", help="Get SRX for a GSM")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("gsm_ids", nargs="+")
    subparser.set_defaults(func=gsm_to_srx)

    # pysradb srp-to-gse
    subparser = subparsers.add_parser("srp-to-gse", help="Get GSE for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [sample_accession, run_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_gse)

    # pysradb srp-to-srr
    subparser = subparsers.add_parser("srp-to-srr", help="Get SRR for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [experiment_accession (SRX),
                                            sample_accession (SRS),
                                            study_alias (GSE),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srr)

    # pysradb srp-to-srs
    subparser = subparsers.add_parser("srp-to-srs", help="Get SRS for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [run_accession (SRR),
                                            study_accession (SRP),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srs)

    # pysradb srp-to-srx
    subparser = subparsers.add_parser("srp-to-srx", help="Get SRX for a SRP")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            run_accession (SRR),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            run_alias (GSM_r)',
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srp_id")
    subparser.set_defaults(func=srp_to_srx)

    # pysradb srr-to-gsm
    subparser = subparsers.add_parser("srr-to-gsm", help="Get GSM for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             study_accession (SRP),
                                             run_alias (GSM_r),
                                             sample_alias (GSM_),
                                             experiment_alias (GSM),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_gsm)

    # pysradb srr-to-srp
    subparser = subparsers.add_parser("srr-to-srp", help="Get SRP for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             sample_accession (SRS),
                                             run_alias (GSM_r),
                                             experiment_alias (GSM),
                                             sample_alias (GSM_),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srp)

    # pysradb srr-to-srs
    subparser = subparsers.add_parser("srr-to-srs", help="Get SRS for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""'Output additional columns: [experiment_accession (SRX),
                                             study_accession (SRP),
                                             run_alias (GSM_r),
                                             sample_alias (GSM_),
                                             experiment_alias (GSM),
                                             study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srs)

    # pysradb srr-to-srx
    subparser = subparsers.add_parser("srr-to-srx", help="Get SRX for a SRR")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [sample_accession (SRS),
                                            study_accession (SRP),
                                            run_alias (GSM_r),
                                            experiment_alias (GSM),
                                            sample_alias (GSM_),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srr_ids", nargs="+")
    subparser.set_defaults(func=srr_to_srx)

    # pysradb srs-to-gsm
    subparser = subparsers.add_parser("srs-to-gsm", help="Get GSM for a SRS")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srs_ids", nargs="+")
    subparser.set_defaults(func=srs_to_gsm)

    # pysradb srs-to-srx
    subparser = subparsers.add_parser("srs-to-srx", help="Get SRX for a SRS")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srs_ids", nargs="+")
    subparser.set_defaults(func=srs_to_srx)

    # pysradb srx-to-srp
    subparser = subparsers.add_parser("srx-to-srp", help="Get SRP for a SRX")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="""Output additional columns: [run_accession (SRR),
                                            sample_accession (SRS),
                                            experiment_alias (GSM),
                                            run_alias (GSM_r),
                                            sample_alias (GSM),
                                            study_alias (GSE)]""",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srp)

    # pysradb srx-to-srr
    subparser = subparsers.add_parser("srx-to-srr", help="Get SRR for a SRX")
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [sample_accession, study_accession]",
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srr)

    # pysradb srx-to-srs
    subparser = subparsers.add_parser("srx-to-srs", help="Get SRS for a SRX")
    subparser.add_argument("--saveto", help="Save output to file")
    subparser.add_argument(
        "--detailed",
        action="store_true",
        help="Output additional columns: [run_accession, study_accession]",
    )
    subparser.add_argument(
        "--desc", action="store_true", help="Should sample_attribute be included"
    )
    subparser.add_argument(
        "--expand", action="store_true", help="Should sample_attribute be expanded"
    )
    subparser.add_argument("srx_ids", nargs="+")
    subparser.set_defaults(func=srx_to_srs)

    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])
    if args.command == "metadata":
        metadata(
            args.srp_id,
            args.assay,
            args.desc,
            args.detailed,
            args.expand,
            args.saveto,
        )
    elif args.command == "download":
        download(
            args.out_dir,
            args.srx,
            args.srp,
            args.geo,
            args.skip_confirmation,
            args.col,
            args.use_ascp,
            args.threads,
        )
    elif args.command == "search":
        flags = vars(args)
        if flags.pop("geo_info"):
            get_geo_search_info()
        else:
            search(
                flags.pop("saveto"),
                flags.pop("db"),
                flags.pop("verbosity"),
                flags.pop("max"),
                flags,
            )
    elif args.command == "gse-to-gsm":
        gse_to_gsm(args.gse_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gse-to-srp":
        gse_to_srp(args.gse_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-gse":
        gsm_to_gse(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srp":
        gsm_to_srp(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srr":
        gsm_to_srr(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srs":
        gsm_to_srs(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "gsm-to-srx":
        gsm_to_srx(args.gsm_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-gse":
        srp_to_gse(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srr":
        srp_to_srr(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srs":
        srp_to_srs(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srp-to-srx":
        srp_to_srx(args.srp_id, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-gsm":
        srr_to_gsm(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srp":
        srr_to_srp(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srs":
        srr_to_srs(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srr-to-srx":
        srr_to_srx(args.srr_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srs-to-gsm":
        srs_to_gsm(args.srs_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srs-to-srx":
        srs_to_srx(args.srs_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srp":
        srx_to_srp(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srr":
        srx_to_srr(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)
    elif args.command == "srx-to-srs":
        srx_to_srs(args.srx_ids, args.saveto, args.detailed, args.desc, args.expand)


if __name__ == "__main__":
    parse_args(sys.argv[1:])



"""Utilities to interact with GEO online"""

import gzip
import os
import re
import requests
import sys
from lxml import html

from .download import download_file
from .geodb import GEOdb
from .utils import _get_url
from .utils import copyfileobj
from .utils import get_gzip_uncompressed_size

PY3 = True
if sys.version_info[0] < 3:
    PY3 = False


class GEOweb(GEOdb):
    def __init__(self):
        """Initialize GEOweb without any database."""

    def get_download_links(self, gse):
        """Obtain all links from the GEO FTP page.

        Parameters
        ----------
        gse: string
             GSE ID

        Returns
        -------
        links: list
               List of all valid downloadable links present for a GEO ID
        """
        prefix = gse[:-3]
        url = f"https://ftp.ncbi.nlm.nih.gov/geo/series/{prefix}nnn/{gse}/suppl/"
        link_objects = html.fromstring(requests.get(url).content).xpath("//a")
        links = [i.attrib["href"] for i in link_objects]
        # remove vulnerability link
        links = [
            link
            for link in links
            if link != "https://www.hhs.gov/vulnerability-disclosure-policy/index.html"
        ]
        # Check if returned results are a valid page - a link to the
        # home page only exists where the GSE ID dow not exist
        if "/" in links:
            raise KeyError(f"The provided GEO ID {gse} does not exist.")

        # The list of links for a valid GSE ID also contains a link to
        # the parent directory - we do not want that
        links = [i for i in links if "geo/series/" not in i]

        # The links are relative, we need absolute links to download
        links = [i for i in links]

        return links, url

    def download(self, links, root_url, gse, verbose=False, out_dir=None):
        """Download GEO files.

        Parameters
        ----------
        links: list
               List of all links valid downloadable present for a GEO ID
        root_url: string
                  url for root directory for a GEO ID
        gse: string
             GEO ID
        verbose: bool
                 Print file list
        out_dir: string
                 Directory location for download
        """
        if out_dir is None:
            out_dir = os.path.join(os.getcwd(), "pysradb_downloads")

        # store output in a separate directory
        out_dir = os.path.join(out_dir, gse)
        os.makedirs(out_dir, exist_ok=True)

        # Display files to be downloaded
        print("\nThe following files will be downloaded: \n")
        for link in links:
            print(link)
        print(os.linesep)
        # Check if we can access list of files in the tar file
        tar_list = [i for i in links if ".tar" in i]
        if "filelist.txt" in links:
            tar_file = tar_list[0]
            if verbose:
                print(f"\nThe tar file {tar_file} contains the following files:\n")
                file_list_contents = requests.get(
                    root_url + "filelist.txt"
                ).content.decode("utf-8")
                print(file_list_contents)

        # Download files
        for link in links:
            # add a prefix to distinguish filelist.txt from different downloads
            prefix = ""
            if link == "filelist.txt":
                prefix = gse + "_"
            geo_path = os.path.join(out_dir, prefix + link)
            download_file(
                root_url.lstrip("https://") + link, geo_path, show_progress=True
            )



"""Tests for GEOweb"""

import os
import time

import pandas as pd
import pytest

from pysradb.geoweb import GEOweb


@pytest.fixture(scope="module")
def geoweb_connection():
    db = GEOweb()
    time.sleep(2)
    return db


def test_valid_download_links(geoweb_connection):
    """Test if all links for a project are scraped"""
    links, url = geoweb_connection.get_download_links("GSE161707")
    assert links == ["GSE161707_RAW.tar", "filelist.txt"]


def test_invalid_download_links(geoweb_connection):
    """Test if invalid GEO ID raises the expected error"""
    with pytest.raises(KeyError):
        links, url = geoweb_connection.get_download_links("GSE1691709")


def test_file_download(geoweb_connection):
    """Test if file actually gets downloaded"""
    geoweb_connection.download(
        links=["GSE161707_RAW.tar", "filelist.txt"],
        root_url="https://ftp.ncbi.nlm.nih.gov/geo/series/GSE161nnn/GSE161707/suppl/",
        gse="GSE161707",
        out_dir="geoweb_downloads",
    )
    assert os.path.getsize("geoweb_downloads/GSE161707/GSE161707_RAW.tar")
    assert os.path.getsize("geoweb_downloads/GSE161707/GSE161707_filelist.txt")



"""Tests for utils.py
"""

import pytest

from pysradb.utils import *


@pytest.fixture(scope="module")
def invalid_name():
    return "Red blood cells"


@pytest.fixture(scope="module")
def valid_name():
    return "Homo sapiens"


def invalid_scientific_name_to_taxid(invalid_name):
    with pytest.raises(IncorrectFieldException) as e:
        scientific_name_to_taxid(invalid_name)
    assert "Unknown scientific name" in str(e.value)


def valid_scientific_name_to_taxid(valid_name):
    assert scientific_name_to_taxid(valid_name) == "9606"



"""Tests for SRAweb"""

import time

import pandas as pd
import pytest

from pysradb.sraweb import SRAweb


@pytest.fixture(scope="module")
def sraweb_connection():
    db = SRAweb()
    time.sleep(2)
    return db


def test_sra_metadata(sraweb_connection):
    """Test if metadata has right number of entries"""
    df = sraweb_connection.sra_metadata("SRP016501")
    assert df.shape[0] == 134


def test_sra_metadata_missing_orgname(sraweb_connection):
    """Test if metadata has right number of entries"""
    df = sraweb_connection.sra_metadata("ERP000171")
    # See: https://github.com/saketkc/pysradb/issues/46#issuecomment-657268760
    assert sum(pd.isna(df.organism_name.tolist())) > 0


def test_sra_metadata_multiple(sraweb_connection):
    """Test if metadata has right number of entries"""
    df = sraweb_connection.sra_metadata(["SRP016501", "SRP096025", "SRP103009"])
    assert list(sorted(df.study_accession.unique())) == [
        "SRP016501",
        "SRP096025",
        "SRP103009",
    ]


def test_sra_metadata_multiple_detailed(sraweb_connection):
    """Test if metadata has right number of entries"""
    df = sraweb_connection.sra_metadata(["SRP002605", "SRP098789"], detailed=True)
    columns = ["treatment time", "library type", "transfection", "time"]
    assert len(set(columns).intersection(set(df.columns))) == 4
    ftp_cols = [
        "ena_fastq_http",
        "ena_fastq_http_1",
        "ena_fastq_http_2",
        "ena_fastq_ftp",
        "ena_fastq_ftp_1",
        "ena_fastq_ftp_2",
    ]
    assert len(set(ftp_cols).intersection(set(df.columns))) == 6


def test_tissue_column(sraweb_connection):
    """Test if tissue column exists"""
    df = sraweb_connection.sra_metadata("SRP096025", detailed="True")
    assert list(df["tissue"]) == ["Kidney"] * 4


def test_metadata_exp_accession(sraweb_connection):
    """Test if experiment_accession column is correct"""
    df = sraweb_connection.sra_metadata("SRP103009", detailed="True")
    assert "SRX2705123" in list(df["experiment_accession"])


def test_fetch_gds_results(sraweb_connection):
    """Test if fetch_gds_result returns correct values"""
    df = sraweb_connection.fetch_gds_results("GSE34438")
    assert df["accession"][1] == "GSM849112"


def test_srp_to_gse(sraweb_connection):
    """Test if srp is converted to gse correctly"""
    df = sraweb_connection.srp_to_gse("SRP009836")
    assert df["study_alias"][0] == "GSE34438"


def test_srp_to_srr(sraweb_connection):
    """Test if srp is converted to srr correctly"""
    df = sraweb_connection.srp_to_srr("SRP002605", detailed=True)
    assert df["run_accession"].tolist()[:5] == [
        "SRR057511",
        "SRR057512",
        "SRR057513",
        "SRR057514",
        "SRR057515",
    ]


def test_srp_to_srs(sraweb_connection):
    """Test if srp is converted to srs correctly"""
    df = sraweb_connection.srp_to_srs("SRP014542")
    assert sorted(list(df["sample_accession"])) == [
        "SRS351513",
        "SRS351514",
        "SRS351515",
        "SRS351516",
        "SRS351517",
        "SRS351518",
    ]


def test_srp_to_srx(sraweb_connection):
    """Test if srp is converted to srx correctly"""
    df = sraweb_connection.srp_to_srx("SRP044932")
    assert list(df["experiment_accession"]) == ["SRX663253", "SRX663254"]


def test_gse_to_gsm(sraweb_connection):
    """Test if gse is converted to gsm correctly"""
    df = sraweb_connection.gse_to_gsm("GSE56924", detailed=False)
    assert df.shape[0] == 96


def test_gse_to_gsm1(sraweb_connection):
    """Test if gse_to_gsm works without passing `detailed` parameter"""
    df = sraweb_connection.gse_to_gsm("GSE63858")
    assert list(sorted(df["experiment_alias"])) == ["GSM1558530", "GSM1558531"]


def test_gse_to_srp(sraweb_connection):
    """Test if gse is converted to srp correctly"""
    df = sraweb_connection.gse_to_srp("GSE63858")
    assert df["study_accession"][0] == "SRP050548"


def test_gsm_to_srp(sraweb_connection):
    """Test if gsm is converted to srp correctly"""
    df = sraweb_connection.gsm_to_srp("GSM1371490")
    assert df["study_accession"][0] == "SRP041298"


def test_gsm_to_gse(sraweb_connection):
    """Test if gsm is converted to gse correctly"""
    df = sraweb_connection.gsm_to_gse("GSM1371490")
    assert df["study_alias"][0] == "GSE56924"


def test_gsm_to_srr(sraweb_connection):
    """Test if gsm is converted to srr correctly"""
    df = sraweb_connection.gsm_to_srr("GSM1371489")
    assert df["run_accession"][0] == "SRR1257271"


def test_gsm_to_srs(sraweb_connection):
    """Test if gsm is converted to srs correctly"""
    df = sraweb_connection.gsm_to_srs("GSM1371469")
    assert df["sample_accession"][0] == "SRS594838"


def test_gsm_to_srx(sraweb_connection):
    """Test if gsm is converted to srx correctly"""
    df = sraweb_connection.gsm_to_srx("GSM1371454")
    assert list(df["experiment_accession"]) == ["SRX522468"]


def test_srr_to_gsm(sraweb_connection):
    df = sraweb_connection.srr_to_gsm("SRR057515")
    assert df["experiment_alias"].tolist()[0] == "GSM546921"


def test_srr_to_srp(sraweb_connection):
    """Test if srr is converted to srp correctly"""
    df = sraweb_connection.srr_to_srp("SRR057511", detailed=False)
    assert list(df["study_accession"]) == ["SRP002605"]


def test_srr_to_srp1(sraweb_connection):
    """Test if srr_to_srp works without passing the `detailed` parameter"""
    df = sraweb_connection.srr_to_srp("SRR057515")
    assert list(df["study_accession"]) == ["SRP002605"]


def test_srr_to_srs(sraweb_connection):
    """Test if srr is converted to srs correctly"""
    df = sraweb_connection.srr_to_srs("SRR057513")
    assert list(df["sample_accession"]) == ["SRS079386"]


def test_srr_to_srx(sraweb_connection):
    """Test if srr is converted to srx correctly"""
    df = sraweb_connection.srr_to_srx("SRR057514")
    assert list(df["experiment_accession"]) == ["SRX021967"]


def test_srs_to_gsm(sraweb_connection):
    """Test if srs is converted to gsm correctly"""
    df = sraweb_connection.srs_to_gsm("SRS079386")
    assert list(df["experiment_alias"]) == ["GSM546921"]


def test_srs_to_srx(sraweb_connection):
    """Test if srs is converted to srx correctly"""
    df = sraweb_connection.srs_to_srx("SRS594838")
    assert list(df["experiment_accession"]) == ["SRX522483"]


def test_srx_to_gsm(sraweb_connection):
    """Test if srx is converted to gsm correctly"""
    df = sraweb_connection.srx_to_gsm("SRX663253")
    assert list(df["experiment_alias"]) == ["GSM1446832"]


def test_srx_to_srp(sraweb_connection):
    """Test if srx is converted to srp correctly"""
    df = sraweb_connection.srx_to_srp("SRX663254")
    assert list(df["study_accession"]) == ["SRP044932"]


def test_srx_to_srr(sraweb_connection):
    """Test if srx is converted to srr correctly"""
    df = sraweb_connection.srx_to_srr("SRX2705123")
    assert list(df["run_accession"]) == ["SRR5413172"]


def test_srx_to_srs(sraweb_connection):
    """Test if srx is converted to srs correctly"""
    df = sraweb_connection.srx_to_srs("SRX663253")
    assert list(df["sample_accession"]) == ["SRS668126"]


def test_xmlns_id(sraweb_connection):
    df = sraweb_connection.sra_metadata(["GSM1013144", "GSM2520660"])
    assert list(df["library_layout"]) == ["PAIRED", "SINGLE"]


def test_GCP_url(sraweb_connection):
    df = sraweb_connection.sra_metadata(["SRP002605"], detailed=True)
    assert df["gcp_url"].tolist()[-1].startswith("gs:")


def test_GCP_url2(sraweb_connection):
    df = sraweb_connection.sra_metadata(["DRR138929"], detailed=True)
    assert df["gcp_url"].tolist()[-1].startswith("gs:")



"""This file contains the search classes for the search feature.
"""
import os
import re
import sys
import time
import urllib
import xml.etree.ElementTree as Et
from json import JSONDecodeError

import pandas as pd
import requests
from tqdm import tqdm

from .exceptions import IncorrectFieldException
from .exceptions import MissingQueryException
from .utils import requests_3_retries
from .utils import scientific_name_to_taxid

SEARCH_REQUEST_TIMEOUT = 20
SRA_SEARCH_GROUP_SIZE = 300


class QuerySearch:
    """This is the base class for the search feature.

    This class takes as input the user's search query, which has been
    tokenized by the ArgParser. The query will be sent to either SRA or ENA
    depending on the user's input, and the results will be returned as a
    pandas dataframe.

    Attributes
    ----------
    self.df: Pandas DataFrame
        The search result belonging to this search instance

    Parameters
    ----------
    verbosity : integer
        The level of details of the search result.
    return_max : int
        The maximum number of entries to be returned.
    query : str
        The main query string.
    accession : str
        A relevant study / experiment / sample / run accession number.
    organism  : str
        Scientific name of the sample organism
    layout : str
        Library layout. Possible inputs: single, paired
    mbases : int
        Size of the sample of interest rounded to the nearest megabase.
    publication_date : str
        The publication date of the run in the format dd-mm-yyyy. If a
        date range is desired, input should be in the format of
        dd-mm-yyyy:dd-mm-yyyy
    platform : str
        Sequencing platform used for the run. Some possible inputs include:
        illumina, ion torrent, oxford nanopore
    selection : str
        Library selection. Some possible inputs: cdna, chip, dnase, pcr
    source : str
        Library source. Some possible inputs: genomic, metagenomic,
        transcriptomic
    strategy : str
        Library Preparation strategy. Some possible inputs: wgs, amplicon,
        rna seq
    title : str
        Title of the experiment associated with the run
    suppress_validation: bool
        Defaults to False. If this is set to True, the user input format
        checks will be skipped.
        Setting this to True may cause the program to behave in unexpected
        ways, but allows the user to search queries that does not pass the
        format check.

    Methods
    -------
    get_df()
        Returns the dataframe storing this search result.

    search()
        Executes the search.

    show_result_statistics()
        Shows summary information about search results.

    visualise_results()
        Generate graphs that visualise the search results.

    get_plot_objects():
        Get the plot objects for plots generated.

    """

    def __init__(
        self,
        verbosity=2,
        return_max=20,
        query=None,
        accession=None,
        organism=None,
        layout=None,
        mbases=None,
        publication_date=None,
        platform=None,
        selection=None,
        source=None,
        strategy=None,
        title=None,
        suppress_validation=False,
    ):
        try:
            int_verbosity = int(verbosity)
            if int_verbosity not in range(4):
                raise ValueError
        except (TypeError, ValueError):
            raise IncorrectFieldException(
                f"Incorrect verbosity format: {verbosity}\n"
                "Verbosity must be an integer between 0 to 3 inclusive."
            )
        try:
            int_return_max = int(return_max)
            if int_return_max <= 0:
                raise ValueError
        except (TypeError, ValueError):
            raise IncorrectFieldException(
                f"Incorrect return_max format: {return_max}\n"
                "return_max must be a positive integer."
            )
        self.verbosity = int_verbosity
        self.return_max = int_return_max
        self.fields = {
            "query": query,
            "accession": accession,
            "organism": organism,
            "layout": layout,
            "mbases": mbases,
            "publication_date": publication_date,
            "platform": platform,
            "selection": selection,
            "source": source,
            "strategy": strategy,
            "title": title,
        }
        for k in self.fields:
            if type(self.fields[k]) == list:
                self.fields[k] = " ".join(self.fields[k])
        self.df = pd.DataFrame()
        # Verify that not all query fields are empty
        if not any(self.fields.values()):
            raise MissingQueryException()
        if not suppress_validation:
            self._validate_fields()
        self.stats = {
            "study": "-",
            "experiment": "-",
            "run": "-",
            "sample": "-",
            "Date range": "-",
            "Organisms": "-",
            "Library strategy": "-",
            "Library source": "-",
            "Library selection": "-",
            "Library layout": "-",
            "Platform": "-",
            "count_mean": "-",
            "count_median": "-",
            "count_stdev": "-",
        }
        self.plot_objects = {}

    def _input_multi_regex_checker(self, regex_matcher, input_query, error_message):
        """Checks if the user input match exactly 1 of the possible regex.

        This is a helper method for _validate_fields. It takes as input a
        dictionary of regex expression : accepted string by API pairs, and
        an input string, and verifies that the input string matches exactly
        one of the regex expressions.
        Matching multiple expressions indicates the input string is
        ambiguous, while matching none of the expressions suggests the
        input will likely produce no results or an error.
        Once matched, this method formats the user input so that it
        can be accepted by the API to be queried, as ENA especially expect
        case-sensitive exact matches for search queries.

        Parameters
        ----------
        regex_matcher : dict
            dictionary of regex expression : accepted string by API pairs.
        input_query : str
            input string for a particular query field.
        error_message : str
            error message to be shown if input_query does not match any of
            the regex expressions in regex_matcher.

        Returns
        -------
        tuple
            tuple pair of (input_query, message).
            message is "" if no format error has been identified, error
            message otherwise.
        """
        matched_strings = []
        for regex_expression in regex_matcher:
            if re.match(regex_expression, input_query, re.IGNORECASE):
                matched_strings.append(regex_matcher[regex_expression])
        if not matched_strings:
            return input_query, error_message
        elif len(matched_strings) == 1:
            return matched_strings[0], ""
        else:
            message = (
                f"Multiple potential matches have been identified for {input_query}:\n"
                f"{matched_strings}\n"
                f"Please check your input.\n\n"
            )
            return input_query, message

    def _validate_fields(self):
        """Verifies that user input format is correct.

        This helper function tries to match the input query strings from
        the user to the exact string that is accepted by both SRA and ENA

        Note: as of the implementation of this method, ENA does not have
        a documentation page listing the accepted values for query
        parameters. The list of parameters below were collected from ENA's
        advanced search page: https://www.ebi.ac.uk/ena/browser/advanced-search

        Updating new values:
        If any new values are accepted by ENA, it should appear under
        the corresponding parameter in the page. To update this method,
        think of a regex that captures what the user may type for the new
        value, and include it in the respective xxx_matcher dictionary
        below as a regex:value key value pair.

        Eg: if a new sequencing platform, Pokemon, is added to ENA,
        navigate to "Instrument Platform" parameter on ENA's advanced
        search page and copy the corresponding phrase (eg "poKe_Mon").
        Then add the key value pair ".*poke.*": "poKe_Mon" to
        platform_matcher below.

        Unlike SRA, ENA requires supplied param values to be exact match
        to filter accordingly (eg "cDNA_oligo_dT"), which motivated this
        feature.

        Raises
        ------
        IncorrectFieldException
            If the input to any query field is in the wrong format

        """

        message = ""

        # verify layout
        if self.fields["layout"] and str(self.fields["layout"]).upper() not in [
            "SINGLE",
            "PAIRED",
        ]:
            message += (
                f"Incorrect layout field format: {self.fields['layout']}\n"
                "--layout must be either SINGLE or PAIRED\n\n"
            )
        # verify mbases
        if self.fields["mbases"]:
            try:
                self.fields["mbases"] = int(self.fields["mbases"])
                if self.fields["mbases"] <= 0:
                    raise ValueError
            except (ValueError, TypeError):
                message += (
                    f"Incorrect mbases format: {self.fields['mbases']}\n"
                    f"--mbases must be a positive integer\n\n"
                )
        # verify publication_date
        date_regex = "(0[1-9]|[12][0-9]|3[01])-(0[1-9]|1[012])-(19|20)[0-9]{2}"
        if self.fields["publication_date"] and not re.match(
            f"^{date_regex}(:{date_regex})?$", self.fields["publication_date"]
        ):
            message += (
                f"Incorrect publication date format: {self.fields['publication_date']}\n"
                f"Expected --publication-date format: dd-mm-yyyy or dd-mm-yyyy:dd-mm-yyyy, between 1900-2099\n\n"
            )
        # verify platform
        platform_matcher = {
            ".*oxford.*|.*nanopore.*": "OXFORD_NANOPORE",
            ".*illumina.*": "ILLUMINA",
            ".*ion.*torrent.*": "ION_TORRENT",
            ".*capillary.*": "CAPILLARY",
            ".*pacbio.*|.*smrt.*": "PACBIO_SMRT",
            ".*abi.*solid.*": "ABI_SOLID",
            ".*bgi.*": "BGISEQ",
            ".*454.*": "LS454",
            ".*complete.*genomics.*": "COMPLETE_GENOMICS",
            ".*helicos.*": "HELICOS",
        }
        if self.fields["platform"]:
            error_message = (
                f"Incorrect platform: {self.fields['platform']}\n"
                f"--platform must be one of the following: \n"
                f"OXFORD_NANOPORE, ILLUMINA, ION_TORRENT, \n"
                f"CAPILLARY, PACBIO_SMRT, ABI_SOLID, \n"
                f"BGISEQ, LS454, COMPLETE_GENOMICS, HELICOS\n\n"
            )
            output = self._input_multi_regex_checker(
                platform_matcher, self.fields["platform"], error_message
            )
            if output[1]:
                message += output[1]
            else:
                self.fields["platform"] = output[0]
        # verify selection
        selection_matcher = {
            ".*methylcytidine.*": "5-methylcytidine antibody",
            ".*cage.*": "CAGE",
            r".*chip\s*$": "ChIP",
            ".*chip.*seq.*": "ChIP-Seq",
            ".*dnase.*": "DNase",
            ".*hmpr.*": "HMPR",
            ".*hybrid.*": "Hybrid Selection",
            r".*inverse.*rrna\s*$": "Inverse rRNA",
            ".*inverse.*rrna.*selection.*": "Inverse rRNA selection",
            ".*mbd2.*protein.*methyl.*cpg.*binding.*domain.*": "MBD2 protein methyl-CpG binding domain",
            ".*mda.*": "MDA",
            ".*mf.*": "MF",
            ".*mnase.*": "MNase",
            ".*msll.*": "MSLL",
            r"^\s*oligo.*dt.*": "Oligo-dT",
            r"^\s*pcr\s*$": "PCR",
            ".*poly[ -_]*a.*": "PolyA",
            ".*race.*": "RACE",
            r".*random\s*$": "RANDOM",
            ".*random.*pcr.*": "RANDOM PCR",
            ".*rt[ -_]*pcr.*": "RT-PCR",
            ".*reduced.*representation.*": "Reduced Representation",
            ".*restriction.*digest.*": "Restriction Digest",
            r".*cdna\s*$": "cDNA",
            ".*cdna.*oligo.*dt": "cDNA_oligo_dT.*",  # ENA only
            ".*cdna.*random.*priming": "cDNA_randomPriming.*",  # ENA only
            ".*other.*": "other",
            ".*padlock.*probes.*capture.*method.*": "padlock probes capture method",
            ".*repeat.*fractionation.*": "repeat fractionation",
            ".*size.*fractionation.*": "size fractionation",
            ".*unspecified.*": "unspecified",
        }
        if self.fields["selection"]:
            error_message = (
                f"Incorrect selection: {self.fields['selection']}\n"
                f"--selection must be one of the following: \n"
                f"5-methylcytidine antibody, CAGE, ChIP, ChIP-Seq, DNase, HMPR, Hybrid Selection,  \n"
                f"Inverse rRNA, Inverse rRNA selection, MBD2 protein methyl-CpG binding domain, \n"
                f"MDA, MF, MNase, MSLL, Oligo-dT, PCR, PolyA, RACE, RANDOM, RANDOM PCR, RT-PCR,  \n"
                f"Reduced Representation, Restriction Digest, cDNA, cDNA_oligo_dT, cDNA_randomPriming \n"
                f"other, padlock probes capture method, repeat fractionation, size fractionation, \n"
                f"unspecified\n\n"
            )
            output = self._input_multi_regex_checker(
                selection_matcher, self.fields["selection"], error_message
            )
            if output[1]:
                message += output[1]
            else:
                self.fields["selection"] = output[0]
        # verify source
        source_matcher = {
            r"^\s*genomic\s*$": "GENOMIC",
            ".*genomic.*single.*cell.*": "GENOMIC SINGLE CELL",
            ".*metagenomic.*": "METAGENOMIC",
            ".*metatranscriptomic.*": "METATRANSCRIPTOMIC",
            ".*other.*": "OTHER",
            ".*synthetic.*": "SYNTHETIC",
            r"^\s*transcriptomic\s*$": "TRANSCRIPTOMIC",
            ".*transcriptomic.*single.*cell.*": "TRANSCRIPTOMIC SINGLE CELL",
            ".*viral.*rna.*": "VIRAL RNA",
        }
        if self.fields["source"]:
            error_message = (
                f"Incorrect source: {self.fields['source']}\n"
                f"--source must be one of the following: \n"
                f"GENOMIC, GENOMIC SINGLE CELL, METAGENOMIC,  \n"
                f"METATRANSCRIPTOMIC, OTHER, SYNTHETIC, \n"
                f"TRANSCRIPTOMIC, TRANSCRIPTOMIC SINGLE CELL, VIRAL RNA\n\n"
            )
            output = self._input_multi_regex_checker(
                source_matcher, self.fields["source"], error_message
            )
            if output[1]:
                message += output[1]
            else:
                self.fields["source"] = output[0]
        # verify strategy
        strategy_matcher = {
            ".*amplicon.*": "AMPLICON",
            ".*atac.*": "ATAC-seq",
            ".*bisulfite.*": "Bisulfite-Seq",
            r"^\s*clone\s*$": "CLONE",
            ".*cloneend.*": "CLONEEND",
            ".*cts.*": "CTS",
            ".*chia.*|.*pet.*": "ChIA-PET",
            ".*chip.*seq.*": "ChIP-Seq",
            ".*dnase.*|.*hypersensitivity.*": "DNase-Hypersensitivity",
            r"^\s*est\s*$": "EST",
            ".*faire.*": "FAIRE-seq",
            ".*finishing.*": "FINISHING",
            ".*fl.*cdna.*": "FL-cDNA",
            ".*hi.*c.*": "Hi-C",
            ".*mbd.*": "MBD-Seq",
            ".*mnase.*": "MNase-Seq",
            ".*mre.*": "MRE-Seq",
            ".*medip.*": "MeDIP-Seq",
            ".*other.*": "OTHER",
            ".*poolclone.*": "POOLCLONE",
            ".*rad.*": "RAD-Seq",
            ".*rip.*": "RIP-Seq",
            r"^\s*rna.*seq": "RNA-Seq",
            ".*selex.*": "SELEX",
            ".*synthetic.*|.*long.*read.*": "Synthetic-Long-Read",
            ".*targeted.*capture.*": "Targeted-Capture",
            ".*tethered.*chromatin.*conformation.*capture.*|.*tccc.*": "Tethered Chromatin Conformation Capture",
            ".*tn.*": "Tn-Seq",
            ".*validation.*": "VALIDATION",
            ".*wcs.*": "WCS",
            ".*wga.*": "WGA",
            ".*wgs.*": "WGS",
            ".*wxs.*": "WXS",
            ".*mirna.*": "miRNA-Seq",
            ".*ncrna.*": "ncRNA-Seq",
            ".*ssrna.*": "ssRNA-seq",
            ".*gbs.*": "GBS",
        }
        if self.fields["strategy"]:
            error_message = (
                f"Incorrect strategy: {self.fields['strategy']}\n"
                f"--strategy must be one of the following: \n"
                f"AMPLICON, ATAC-seq, Bisulfite-Seq, CLONE, CLONEEND, CTS, ChIA-PET, ChIP-Seq, \n"
                f"DNase-Hypersensitivity, EST, FAIRE-seq, FINISHING, FL-cDNA, Hi-C, MBD-Seq, MNase-Seq,\n"
                f"MRE-Seq, MeDIP-Seq, OTHER, POOLCLONE, RAD-Seq, RIP-Seq, RNA-Seq, SELEX, \n"
                f"Synthetic-Long-Read, Targeted-Capture, Tethered Chromatin Conformation Capture, \n"
                f"Tn-Seq, VALIDATION, WCS, WGA, WGS, WXS, miRNA-Seq, ncRNA-Seq, ssRNA-seq, GBS\n\n"
            )
            output = self._input_multi_regex_checker(
                strategy_matcher, self.fields["strategy"], error_message
            )
            if output[1]:
                message += output[1]
            else:
                self.fields["strategy"] = output[0]
        if message:
            raise IncorrectFieldException(message)

    def _list_stat(self, stat_header):
        stat = self.stats[stat_header]
        if type(stat) != dict:
            return f"  {stat_header}: {stat}\n"
        keys = sorted(stat.keys())
        stat_breakdown = "\n"
        for key in keys:
            stat_breakdown += f"\t  {key}:  {stat[key]}\n"
        return f"  {stat_header}: {stat_breakdown}\n"

    def show_result_statistics(self):
        """Shows search result statistics."""
        if self.df.empty:
            print(
                "No results are found for the current search query, hence no statistics can be generated."
            )
            return
        stats = (
            "\n  Statistics for the search query:\n"
            + "  =================================\n"
            + f"  Number of unique studies: {self.stats['study']}\n"
            + f"  Number of unique experiments: {self.stats['experiment']}\n"
            + f"  Number of unique runs: {self.stats['run']}\n"
            + f"  Number of unique samples: {self.stats['sample']}\n"
            + f"  Mean base count of samples: {self.stats['count_mean']:.3f}\n"
            + f"  Median base count of samples: {self.stats['count_median']:.3f}\n"
            + f"  Sample base count standard deviation: {self.stats['count_stdev']:.3f}\n"
        )

        # Statistics with categorical breakdowns:
        categorical_stats = (
            "Date range",
            "Organisms",
            "Platform",
            "Library strategy",
            "Library source",
            "Library selection",
            "Library layout",
        )
        for categorical_stat in categorical_stats:
            stats += self._list_stat(categorical_stat)
        print(stats)

    def visualise_results(
        self, graph_types=("all",), show=False, saveto="./search_plots/"
    ):
        """Generate graphs that visualise the search results.

        This method will only work if the optional dependency, matplotlib,
        is installed in the system.

        Parameters
        ----------
        graph_types : tuple
            tuple containing strings representing types of graphs to
            generate.
            Possible strings: all, daterange, organism, source, selection, platform,
            basecount
        saveto : str
            directory name where the generated graphs are saved.

        show : bool
            Whether plotted graphs are immediately shown.
        """
        if self.df.empty:
            print(
                "No results are found for the current search query, hence no graphs can be generated."
            )
            return
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            print(
                "The optional dependency, matplotlib, is not available on the system.\n"
                "matplotlib is required to generate graphs to visualise search results.\n"
                'You can install matplotlib by typing "pip install matplotlib" on the command line.\n'
            )
            return
        plt.rcParams["figure.autolayout"] = True
        if not os.path.isdir(saveto):
            os.mkdir(saveto)
        plots = [
            ("Base Count",),
            ("Publication Date",),
            ("Organism",),
            ("Library Source",),
            ("Library Selection",),
            ("Platform",),
            ("Organism", "Publication Date"),
            ("Library Source", "Publication Date"),
            ("Library Selection", "Publication Date"),
            ("Platform", "Publication Date"),
            ("Library Source", "Organism"),
            ("Library Selection", "Organism"),
            ("Platform", "Organism"),
            ("Library Selection", "Library Source"),
            ("Platform", "Library Source"),
            ("Platform", "Library Selection"),
        ]
        plot_keys = {
            "daterange": "Publication Date",
            "organism": "Organism",
            "source": "Library Source",
            "selection": "Library Selection",
            "platform": "Platform",
            "basecount": "Base Count",
        }
        if "all" not in graph_types:
            selected_plots = []
            for graph_type in graph_types:
                if graph_type not in plot_keys:
                    continue
                for plot in plots:
                    if plot_keys[graph_type] in plot and plot not in selected_plots:
                        selected_plots.append(plot)
            plots = selected_plots
        too_many_organisms = False
        if self.stats["graph_raw"]["Organism"].nunique() > 30:
            print(
                "Too many types of organisms to plot (>30). Showing only top 30 organisms."
            )
            too_many_organisms = True
        for plot in plots:
            self._plot_graph(plt, plot, show, saveto, too_many_organisms)

    def search(self):
        pass

    def get_df(self):
        """Getter for the search result dataframe."""
        return self.df

    def get_plot_objects(self):
        """Get the plot objects for plots generated."""
        return self.plot_objects

    def _plot_graph(self, plt, axes, show, savedir, too_many_organisms):
        """Plots a graph based on data from self.stats

        Parameters
        ----------
        axes: tuple
            tuple containing 1 or 2 strings corresponding to the statistics
            to plot. 1 string: Histogram. 2 string: heat map
        savedir: str
            directory to save to
        show: bool
            whether to call plt.show
        """
        timestamp = time.strftime("%Y-%m-%d %H-%M-%S")
        if (
            "Publication Date" in axes
            and self.stats["graph_raw"]["Publication Date"].nunique() > 30
        ):
            self.stats["graph_raw"]["Publication Date"] = self.stats["graph_raw"][
                "Publication Date"
            ].str[:-3]
        if axes == ("Base Count",):
            count = list(self.stats["count_data"])
            if len(count) == 0:
                return
            title = "Histogram of Base Count"
            plt.figure(figsize=(15, 10))
            plt.hist(count, min(70, len(count)), color="#135c1c", log=True)
            plt.xlabel("Base Count", fontsize=14)
            plt.ylabel("Frequency", fontsize=14)
            plt.xticks(rotation=90)
            plt.title(title, fontsize=18)
            plt.savefig(f"{savedir}{title} {timestamp}.svg")
            self.plot_objects[axes] = plt
        elif len(axes) == 1:
            title = f"Histogram of {axes[0]}"
            data = self.stats["graph_raw"][axes[0]].value_counts()
            if too_many_organisms:
                data = data[:30]
            plt.figure(figsize=(15, 10))
            plt.bar(
                range(len(data.values)),
                data.values,
                tick_label=list(data.index),
                color="#135c1c",
            )
            plt.xticks(rotation=90)
            plt.title(title, fontsize=18)
            plt.xlabel(axes[0], fontsize=14)
            plt.ylabel("Frequency", fontsize=14)
            plt.savefig(f"{savedir}{title} {timestamp}.svg")
            self.plot_objects[axes] = plt
        elif len(axes) == 2:
            title = f"Heatmap of {axes[0]} against {axes[1]}"
            df = self.stats["graph_raw"][list(axes)]
            a = df.groupby([axes[0]]).agg({i: "value_counts" for i in df.columns[1:]})
            a = a.rename(columns={axes[1]: f"{axes[1]}_count"})
            b = a.reset_index(level=list(axes))
            piv = (
                pd.pivot_table(
                    b,
                    values=f"{axes[1]}_count",
                    index=[axes[0]],
                    columns=[axes[1]],
                    fill_value=0,
                    aggfunc="sum",
                    margins=True,
                )
                .sort_values("All", ascending=False)
                .drop("All", axis=1)
                .sort_values("All", ascending=False, axis=1)
                .drop("All")
            )
            if too_many_organisms:
                if axes[0] == "Organism":
                    piv = piv[:30]
                else:
                    piv = piv.iloc[:, :30]
            fig, ax = plt.subplots(figsize=(15, 12))
            im = ax.imshow(piv, cmap="Greens")
            fig.colorbar(im, ax=ax)
            ax.set_title(title, fontsize=18)
            ax.set_xticks(range(len(piv.columns)))
            ax.set_yticks(range(len(piv.index)))
            ax.set_xticklabels(piv.columns, rotation=90)
            ax.set_yticklabels(piv.index)
            ax.set_ylabel(axes[0], fontsize=14)
            ax.set_xlabel(axes[1], fontsize=14)
            ax.get_figure().savefig(f"{savedir}{title} {timestamp}.svg")
            self.plot_objects[axes] = (fig, ax)
        if show:
            plt.show()


class SraSearch(QuerySearch):
    """Subclass of QuerySearch that implements search by querying
    NCBI Entrez API

    Methods
    -------
    search()
        sends the user query via requests to NCBI Entrez API and returns
        search results as a pandas dataframe.

    show_result_statistics()
        Shows summary information about search results.

    visualise_results()
        Generate graphs that visualise the search results.

    get_plot_objects():
        Get the plot objects for plots generated.

    get_uids():
        Get NCBI uids retrieved during this search query.

    _format_query_string()
        formats the input user query into a string

    _format_request()
        formats the request payload

    _format_result(content)
        formats the search query output.

    See Also
    --------
    QuerySearch: Superclass of SraSearch

    """

    def __init__(
        self,
        verbosity=2,
        return_max=20,
        query=None,
        accession=None,
        organism=None,
        layout=None,
        mbases=None,
        publication_date=None,
        platform=None,
        selection=None,
        source=None,
        strategy=None,
        title=None,
        suppress_validation=False,
    ):
        super().__init__(
            verbosity,
            return_max,
            query,
            accession,
            organism,
            layout,
            mbases,
            publication_date,
            platform,
            selection,
            source,
            strategy,
            title,
            suppress_validation,
        )
        self.entries = {}
        self.number_entries = 0
        self.uids = []

    def search(self):
        # Step 1: retrieves the list of uids that satisfies the input
        # search query
        payload = self._format_request()
        try:
            r = requests_3_retries().get(
                "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
                params=payload,
                timeout=SEARCH_REQUEST_TIMEOUT,
            )
            r.raise_for_status()
            self.uids = r.json()["esearchresult"]["idlist"]

            # Step 2: retrieves the detailed information for each uid
            # returned, in groups of SRA_SEARCH_GROUP_SIZE.
            if not self.uids:
                print(
                    f"No results found for the following search query: \n {self.fields}"
                )
                return  # If no queries found, return nothing
            pbar = tqdm(total=len(self.uids))
            for i in range(0, len(self.uids), SRA_SEARCH_GROUP_SIZE):
                current_uids = ",".join(
                    self.uids[i : min(i + SRA_SEARCH_GROUP_SIZE, len(self.uids))]
                )
                pbar.update(min(SRA_SEARCH_GROUP_SIZE, len(self.uids) - i))
                payload2 = {"db": "sra", "retmode": "xml", "id": current_uids}

                r = requests_3_retries().get(
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi",
                    params=payload2,
                    timeout=SEARCH_REQUEST_TIMEOUT,
                    stream=True,
                )
                r.raise_for_status()
                r.raw.decode_content = True
                self._format_response(r.raw)
            pbar.close()
            self._format_result()

        except requests.exceptions.Timeout:
            sys.exit(f"Connection to the server has timed out. Please retry.")
        except requests.exceptions.HTTPError:
            sys.exit(
                f"HTTPError: This is likely caused by an invalid search query: "
                f"\nURL queried: {r.url} \nUser query: {self.fields}"
            )

    def get_uids(self):
        """Get NCBI uids retrieved during this search query.

        Note: There is a chance that some uids retrieved do not appear in
        the search result output (Refer to #88)
        """
        return self.uids

    def _format_query_string(self):
        term = ""
        if self.fields["query"]:
            term += self.fields["query"] + " AND "
        if self.fields["accession"]:
            term += self.fields["accession"] + "[Accession] AND "
        if self.fields["organism"]:
            term += self.fields["organism"] + "[Organism] AND "
        if self.fields["layout"]:
            term += self.fields["layout"] + "[Layout] AND "
        if self.fields["mbases"]:
            term += str(self.fields["mbases"]) + "[Mbases] AND "
        if self.fields["publication_date"]:
            dates = []
            for date in self.fields["publication_date"].split(":"):
                dates.append("/".join(date.split("-")[::-1]))
            term += ":".join(dates) + "[PDAT] AND "
        if self.fields["platform"]:
            term += self.fields["platform"] + "[Platform] AND "
        if self.fields["selection"]:
            term += self.fields["selection"] + "[Selection] AND "
        if self.fields["source"]:
            term += self.fields["source"] + "[Source] AND "
        if self.fields["strategy"]:
            term += self.fields["strategy"] + "[Strategy] AND "
        if self.fields["title"]:
            term += self.fields["title"] + "[Title] AND "
        return term[:-5]  # Removing trailing " AND "

    def _format_request(self):
        payload = {
            "db": "sra",
            "term": self._format_query_string(),
            "retmode": "json",
            "retmax": self.return_max,
        }
        return payload

    def _format_response(self, content):
        field_categories = [
            "EXPERIMENT",
            "SUBMISSION",
            "ORGANISATION",
            "STUDY",
            "SAMPLE",
            "Pool",
            "RUN_SET",
        ]
        for event, elem in Et.iterparse(content):
            if elem.tag == "EXPERIMENT_PACKAGE":
                self.number_entries += 1
            elif elem.tag in field_categories:
                self._parse_entry(elem)
        for field in self.entries:
            if len(self.entries[field]) < self.number_entries:
                self.entries[field] += [""] * (
                    self.number_entries - len(self.entries[field])
                )

    def _format_result(self):
        self.df = pd.DataFrame.from_dict(self.entries).replace(
            r"^\s*$", pd.NA, regex=True
        )
        self.entries.clear()
        if self.df.empty:
            return
        # Tabulate statistics
        self._update_stats()

        columns = list(self.df.columns)
        important_columns = [
            "study_accession",
            "experiment_accession",
            "experiment_title",
            "design_description",
            "sample_taxon_id",
            "sample_scientific_name",
            "experiment_library_strategy",
            "experiment_library_source",
            "experiment_library_selection",
            "sample_accession",
            "sample_alias",
            "experiment_instrument_model",
            "pool_member_spots",
            "run_1_size",
            "run_1_accession",
            "run_1_total_spots",
            "run_1_total_bases",
        ]
        temp_cols = []
        for col in important_columns:
            if col in columns:
                temp_cols.append(col)
                columns.remove(col)
        important_columns = temp_cols
        if self.verbosity <= 1:
            temp_dfs = []
            for col in self.df.columns:
                if re.match("run_[0-9]+_accession", col):
                    temp_df = self.df[[col, "experiment_title"]]
                    temp_df = temp_df[~pd.isna(temp_df[col])].rename(
                        columns={
                            col: "run_accession",
                            "experiment_title": "experiment_title",
                        }
                    )
                    temp_dfs.append(temp_df)
            run_dataframe = pd.concat(temp_dfs)
            run_dataframe.sort_values(by=["run_accession"], kind="mergesort")
            self.df = run_dataframe
        if self.verbosity == 0:
            self.df = self.df[["run_accession"]]
        elif self.verbosity == 1:
            pass  # df has already been formatted above
        elif self.verbosity == 2:
            self.df = self.df[important_columns]
        elif self.verbosity == 3:
            self.df = self.df[important_columns + sorted(columns)]
        self.df.dropna(how="all")

    def _parse_entry(self, entry_root):
        """Parses a subset of the XML tree from request stream

        Parameters
        ----------
        entry_root: ElementTree.Element
            root element of the xml tree from requests stream
        """
        field_header = entry_root.tag.lower()
        run_count = 0

        # root element attributes
        for k, v in entry_root.attrib.items():
            self._update_entry(f"{field_header}_{k}".lower(), v)
        for child in entry_root:
            # "*_REF" tags contain duplicate information that can be found
            # somewhere in the xml entry and are skipped
            if child.tag.endswith("_REF"):
                continue
            # IDENTIFIERS contain two types of children tags:
            # PRIMARY_ID, which repeats the accession number and
            # EXTERNAL_ID tags, each containing an alternative ID that is
            # typically used in another database (GEO, ENA etc)
            # IDs are numbered from 1 to differentiate between them.
            elif child.tag == "IDENTIFIERS":
                id_index = 1
                for identifier in child:
                    if identifier.tag == "EXTERNAL_ID":
                        self._update_entry(
                            f"{field_header}_external_id_{id_index}",
                            identifier.text,
                        )
                        self._update_entry(
                            f"{field_header}_external_id_{id_index}_namespace",
                            identifier.get("namespace"),
                        )
            # "*_LINKS" tags contain 0 or more "*_LINK" children tags,
            # each containing information (values) regarding the link
            # Links are numbered from 1 to differentiate between multiple
            # links.
            elif child.tag.endswith("_LINKS"):
                link_index = 1
                for link in child:
                    # Link type. Eg: URL_link, Xref_link
                    self._update_entry(
                        f"{link.tag}_{link_index}_type".lower(),
                        link[0].tag,
                    )
                    # Link values in the form of tag: value.
                    # Eg: label: GEO sample
                    link_value_index = 1
                    for link_value in link[0]:
                        self._update_entry(
                            f"{link.tag}_{link_index}_value_{link_value_index}".lower(),
                            f"{link_value.tag}: {link_value.text}",
                        )
                        link_value_index += 1
                    link_index += 1
            # "*_ATTRIBUTES" tags contain tag - value pairs providing
            # additional information for the Experiment/Sample/Study
            # Attributes are numbered from 1 to differentiate between
            # multiple attributes.
            elif child.tag.endswith("_ATTRIBUTES"):
                attribute_index = 1
                for attribute in child:
                    for val in attribute:
                        self._update_entry(
                            f"{child.tag}_{attribute_index}_{val.tag}".lower(),
                            val.text,
                        )
                    attribute_index += 1
            # Differentiating between sample title and experiment title.
            elif child.tag == "TITLE":
                self._update_entry(f"{field_header}_title", child.text)
            # Parsing platfrom information
            elif child.tag == "PLATFORM":
                platform = child[0]
                self._update_entry("experiment_platform", platform.tag)
                self._update_entry(
                    "experiment_instrument_model",
                    platform[0].text,
                )
            # Parsing individual run information
            elif child.tag == "RUN":
                run_count += 1
                # run attributes
                for k, v in child.attrib.items():
                    self._update_entry(f"run_{run_count}_{k}".lower(), v)
                for elem in child:
                    if elem.tag == "SRAFiles":
                        srafile_index = 1
                        for srafile in elem:
                            for k, v in srafile.attrib.items():
                                self._update_entry(
                                    f"run_{run_count}_srafile_{srafile_index}_{k}".lower(),
                                    v,
                                )
                            alternatives_index = 1
                            for alternatives in srafile:
                                for k, v in alternatives.attrib.items():
                                    self._update_entry(
                                        f"run_{run_count}_srafile_{srafile_index}_alternative_{alternatives_index}_{k}".lower(),
                                        v,
                                    )
                                alternatives_index += 1
                            srafile_index += 1
                    elif elem.tag == "CloudFiles":
                        cloudfile_index = 1
                        for cloudfile in elem:
                            for k, v in cloudfile.attrib.items():
                                self._update_entry(
                                    f"run_{run_count}_cloudfile_{cloudfile_index}_{k}".lower(),
                                    v,
                                )
                            cloudfile_index += 1
                    elif elem.tag == "Bases":
                        for k, v in elem.attrib.items():
                            self._update_entry(
                                f"run_{run_count}_total_base_{k}".lower(),
                                v,
                            )
                        for base in elem:
                            self._update_entry(
                                f"run_{run_count}_base_{base.attrib['value']}_count",
                                base.attrib["count"],
                            )
                    elif elem.tag == "Databases":
                        database_index = 1
                        for database in elem:
                            self._update_entry(
                                f"run_{run_count}_database_{database_index}".lower(),
                                Et.tostring(database).decode(),
                            )
                            database_index += 1
            else:
                for elem in child.iter():
                    # Tags to ignore to avoid confusion
                    if elem.tag in ["PRIMARY_ID", "SINGLE", "PAIRED"]:
                        continue
                    elif elem.text:
                        self._update_entry(
                            f"{field_header}_{elem.tag.lower()}",
                            elem.text,
                        )
                    elif elem.attrib:
                        for k, v in elem.attrib.items():
                            self._update_entry(
                                f"{field_header}_{elem.tag}_{k}".lower(),
                                v,
                            )
            # Parsing library layout (single, paired)
            if field_header == "experiment":
                library_layout = child.find("./LIBRARY_DESCRIPTOR/LIBRARY_LAYOUT")
                if library_layout:
                    library_layout = library_layout[0]
                    self._update_entry(f"library_layout", library_layout.tag)
                    # If library layout is paired, information such as nominal
                    # standard deviation and length, etc are provided as well.
                    if library_layout.tag == "PAIRED":
                        for k, v in library_layout.attrib.items():
                            self._update_entry(
                                f"library_layout_{k}".lower(),
                                v,
                            )

    def _update_entry(self, field_name, field_content):
        """Adds information from a field into the entries dictionary

        This is a helper function that adds information parsed from the XML
        output from SRA into a dictionary of lists, for easier conversion
        into a Pandas dataframe later. Dictionary key is created if it
        doesn't exist yet. For entries that does not have information
        belonging to a field, the corresponding list will be padded with
        empty strings.

        Parameters
        ----------
        field_name: str
            Name of the field where a value belonging to an entry is to be
            added
        field_content: str
            Value to be added
        """
        if field_name not in self.entries:
            self.entries[field_name] = []
        if len(self.entries[field_name]) > self.number_entries:
            return
        self.entries[field_name] += [""] * (
            self.number_entries - len(self.entries[field_name])
        ) + [field_content]

    def _update_stats(self):
        # study
        self.stats["study"] = self.df["study_accession"].nunique()
        # experiment
        self.stats["experiment"] = self.df["experiment_accession"].nunique()
        # run
        runs = self._merge_selected_columns(r"^run.*accession$")
        if not runs.empty:
            self.stats["run"] = runs.nunique()
        # sample
        samples = self._merge_selected_columns(r"^sample.*accession$")
        if not samples.empty:
            self.stats["sample"] = samples.nunique()
        # date range
        daterange = self._merge_selected_columns(r"^run_1_published$")
        if not daterange.empty:
            dates = pd.to_datetime(daterange).dt.to_period("M").astype(str)
            self.stats["Date range"] = dates.value_counts().to_dict()
        # organisms
        organisms = self._merge_selected_columns(r"^sample.*scientific_name.*")
        if not organisms.empty:
            self.stats["Organisms"] = organisms.value_counts().to_dict()
        # strategy
        if "experiment_library_strategy" in self.df.columns:
            self.stats["Library strategy"] = (
                self.df["experiment_library_strategy"].value_counts().to_dict()
            )
        # source
        if "experiment_library_source" in self.df.columns:
            self.stats["Library source"] = (
                self.df["experiment_library_source"].value_counts().to_dict()
            )
        # selection
        if "experiment_library_selection" in self.df.columns:
            self.stats["Library selection"] = (
                self.df["experiment_library_selection"].value_counts().to_dict()
            )
        # layout
        if "library_layout" in self.df.columns:
            self.stats["Library layout"] = (
                self.df["library_layout"].value_counts().to_dict()
            )
        # platform
        if "experiment_platform" in self.df.columns:
            self.stats["Platform"] = (
                self.df["experiment_platform"].value_counts().to_dict()
            )
        # count
        count = self._merge_selected_columns(r"^run_.*_total_bases$").astype("int64")
        self.stats["count_data"] = count
        self.stats["count_mean"] = count.mean()
        self.stats["count_median"] = count.median()
        self.stats["count_stdev"] = count.std()

        # for graphing
        self.stats["graph_raw"] = self.df[
            [
                "sample_scientific_name",
                "experiment_library_strategy",
                "experiment_library_source",
                "experiment_library_selection",
                "run_1_published",
                "experiment_platform",
            ]
        ].rename(
            columns={
                "sample_scientific_name": "Organism",
                "experiment_library_strategy": "Library Strategy",
                "experiment_library_source": "Library Source",
                "experiment_library_selection": "Library Selection",
                "run_1_published": "Publication Date",
                "experiment_platform": "Platform",
            }
        )
        self.stats["graph_raw"]["Publication Date"] = (
            pd.to_datetime(
                self.stats["graph_raw"]["Publication Date"].replace(pd.NA, None)
            )
            .dt.to_period("M")
            .astype(str)
        )

    def _merge_selected_columns(self, regex):
        columns = list(self.df.filter(regex=regex, axis=1).columns)
        if not columns:
            series = pd.Series(dtype="object")
        elif len(columns) == 1:
            series = self.df[columns[0]]
        else:
            series = self.df[columns[0]].append(
                [self.df[c] for c in columns[1:]], ignore_index=True
            )
        return series[~pd.isna(series)]


class EnaSearch(QuerySearch):
    """Subclass of QuerySearch that implements search via querying ENA API


    Methods
    -------
    search()
        sends the user query via requests to ENA API and stores search
        result as an instance attribute in the form of a pandas dataframe

    show_result_statistics()
        Shows summary information about search results.

    visualise_results()
        Generate graphs that visualise the search results.

    get_plot_objects():
        Get the plot objects for plots generated.

    _format_query_string()
        formats the input user query into a string

    _format_request()
        formats the request payload

    _format_result(content)
        formats the search query output and converts it into a pandas
        dataframe

    See Also
    --------
    QuerySearch: Superclass of EnaSearch

    """

    def search(self):
        # This ensures that the spaces in the query string are not
        # converted to '+' by requests.
        payload = urllib.parse.urlencode(
            self._format_request(), quote_via=urllib.parse.quote
        )
        try:
            r = requests_3_retries().get(
                "https://www.ebi.ac.uk/ena/portal/api/search",
                params=payload,
                timeout=SEARCH_REQUEST_TIMEOUT,
            )
            r.raise_for_status()
            self._format_result(r.json())
        except requests.exceptions.Timeout:
            sys.exit(f"Connection to the server has timed out. Please retry.")
        except requests.exceptions.HTTPError:
            sys.exit(
                f"HTTPError: This is likely caused by an invalid search query: "
                f"\nURL queried: {r.url} \nUser query: {self.fields}"
            )
        except JSONDecodeError:
            print(f"No results found for the following search query: \n {self.fields}")
            return  # no results found

    def _format_query_string(self):
        term = ""
        if self.fields["query"]:
            term += rf'(experiment_title="*{self.fields["query"]}*"'
            if not self.fields["accession"]:
                self.fields["query"] = self.fields["query"].upper()
                term += (
                    rf' OR study_accession="{self.fields["query"]}" OR '
                    rf'secondary_study_accession="{self.fields["query"]}" OR '
                    rf'sample_accession="{self.fields["query"]}" OR '
                    rf'secondary_sample_accession="{self.fields["query"]}" OR '
                    rf'experiment_accession="{self.fields["query"]}" OR '
                    rf'submission_accession="{self.fields["query"]}" OR '
                    rf'run_accession="{self.fields["query"]}"'
                )
            term += ") AND "
        if self.fields["accession"]:
            self.fields["accession"] = self.fields["accession"].upper()
            term += (
                rf'(study_accession="{self.fields["accession"]}" OR '
                rf'secondary_study_accession="{self.fields["accession"]}" OR '
                rf'sample_accession="{self.fields["accession"]}" OR '
                rf'secondary_sample_accession="{self.fields["accession"]}" OR '
                rf'experiment_accession="{self.fields["accession"]}" OR '
                rf'submission_accession="{self.fields["accession"]}" OR '
                rf'run_accession="{self.fields["accession"]}") AND '
            )
        if self.fields["organism"]:
            term += rf'tax_eq({scientific_name_to_taxid(self.fields["organism"])}) AND '
        if self.fields["layout"]:
            term += rf'library_layout="{self.fields["layout"].upper()}" AND '
        if self.fields["mbases"]:
            if type(self.fields["mbases"]) != int:
                raise IncorrectFieldException(
                    f"Incorrect mbases format: {self.fields['mbases']}\n"
                    f"--mbases must be an integer"
                )
            upper_limit = self.fields["mbases"] * 1000000 + 500000
            lower_limit = self.fields["mbases"] * 1000000 - 500000
            term += rf"base_count>={lower_limit} AND base_count<{upper_limit} AND "
        if self.fields["publication_date"]:
            dates = self.fields["publication_date"].split(":")
            for i in range(len(dates)):
                dates[i] = "-".join(dates[i].split("-")[::-1])
            if len(dates) == 1:
                term += rf"first_created={dates[0]} AND "
            elif len(dates) == 2:
                term += rf"first_created>={dates[0]} AND first_created<={dates[1]} AND "
            else:
                raise IncorrectFieldException(
                    f"Incorrect publication date format: {self.fields['publication_date']}\n"
                    f"Expected format: dd-mm-yyyy or dd-mm-yyyy:dd-mm-yyyy"
                )
        if self.fields["platform"]:
            term += rf'instrument_platform="{self.fields["platform"]}" AND '
        if self.fields["selection"]:
            term += rf'library_selection="{self.fields["selection"]}" AND '
        if self.fields["source"]:
            term += rf'library_source="{self.fields["source"]}" AND '
        if self.fields["strategy"]:
            term += rf'library_strategy="{self.fields["strategy"]}" AND '
        if self.fields["title"]:
            term += rf'experiment_title="*{self.fields["title"]}*" AND '
        return term[:-5]  # Removing trailing " AND "

    def _format_request(self):
        # Note: ENA's API does not support searching a query in all fields.
        # Currently, if the user does not specify a query field, the query will
        # be matched to experiment_title (aka description),
        # or one of the accession fields
        stats_columns = ()
        payload = {
            "query": self._format_query_string(),
            "result": "read_run",
            "format": "json",
            "limit": self.return_max,
        }

        # Selects the fields to return at different verbosity levels
        if self.verbosity < 3:
            payload["fields"] = (
                "study_accession,"
                "experiment_accession,"
                "experiment_title,"
                "description,"
                "tax_id,"
                "scientific_name,"
                "library_strategy,"
                "library_source,"
                "library_selection,"
                "sample_accession,"
                "sample_title,"
                "instrument_model,"
                "run_accession,"
                "read_count,"
                "base_count,"
                "first_public,"
                "library_layout,"
                "instrument_platform"
            )
        elif self.verbosity == 3:
            payload["fields"] = "all"
        return payload

    def _format_result(self, content):
        if not content:
            return
        self.df = pd.DataFrame.from_dict(content).replace(r"^\s*$", pd.NA, regex=True)

        # Tabulate statistics
        self._update_stats()

        important_columns = [
            "study_accession",
            "experiment_accession",
            "experiment_title",
            "description",
            "tax_id",
            "scientific_name",
            "library_strategy",
            "library_source",
            "library_selection",
            "sample_accession",
            "sample_title",
            "instrument_model",
            "run_accession",
            "read_count",
            "base_count",
        ]

        if self.verbosity == 0:
            self.df = self.df[["run_accession"]]
        elif self.verbosity == 1:
            self.df = self.df[["run_accession", "description"]]
        elif self.verbosity == 2:
            self.df = self.df[important_columns]
        elif self.verbosity == 3:
            columns = list(self.df.columns)
            columns = important_columns + sorted(
                [col for col in columns if col not in important_columns]
            )
            self.df = self.df[columns]
        self.df.dropna(how="all")

    def _update_stats(self):
        # study
        self.stats["study"] = self.df["study_accession"].nunique()
        # experiment
        self.stats["experiment"] = self.df["experiment_accession"].nunique()
        # run
        self.stats["run"] = self.df["run_accession"].nunique()
        # sample
        self.stats["sample"] = self.df["sample_accession"].nunique()
        # date range
        daterange = self.df["first_public"]
        daterange = daterange[~pd.isna(daterange)]
        if not daterange.empty:
            dates = pd.to_datetime(daterange).dt.to_period("M").astype(str)
            self.stats["Date range"] = dates.value_counts().to_dict()
        # organisms
        organisms = self.df["scientific_name"]
        self.stats["Organisms"] = organisms.value_counts().to_dict()
        # strategy
        if "library_strategy" in self.df.columns:
            self.stats["Library strategy"] = (
                self.df["library_strategy"].value_counts().to_dict()
            )
        # source
        if "library_source" in self.df.columns:
            self.stats["Library source"] = (
                self.df["library_source"].value_counts().to_dict()
            )
        # selection
        if "library_selection" in self.df.columns:
            self.stats["Library selection"] = (
                self.df["library_selection"].value_counts().to_dict()
            )
        # layout
        if "library_layout" in self.df.columns:
            self.stats["Library layout"] = (
                self.df["library_layout"].value_counts().to_dict()
            )
        # platform
        if "instrument_platform" in self.df.columns:
            self.stats["Platform"] = (
                self.df["instrument_platform"].value_counts().to_dict()
            )
        # count
        count = self.df["base_count"].copy()
        count = count[~pd.isna(count)].astype("int64")
        self.stats["count_data"] = count
        self.stats["count_mean"] = count.mean()
        self.stats["count_median"] = count.median()
        self.stats["count_stdev"] = count.std()

        # For graphing
        self.stats["graph_raw"] = self.df[
            [
                "scientific_name",
                "library_strategy",
                "library_source",
                "library_selection",
                "first_public",
                "instrument_platform",
            ]
        ].rename(
            columns={
                "scientific_name": "Organism",
                "library_strategy": "Library Strategy",
                "library_source": "Library Source",
                "library_selection": "Library Selection",
                "first_public": "Publication Date",
                "instrument_platform": "Platform",
            }
        )
        self.stats["graph_raw"]["Publication Date"] = (
            pd.to_datetime(self.stats["graph_raw"]["Publication Date"])
            .dt.to_period("M")
            .astype(str)
        )


class GeoSearch(SraSearch):
    """Subclass of SraSearch that can query both GEO DataSets and SRA API.

    Methods
    -------
    search()
        sends the user query via requests to SRA, GEO DataSets, or both
        depending on the search query. If query is sent to both APIs,
        the intersection of the two sets of query results are returned.

    show_result_statistics()
        Shows summary information about search results.

    visualise_results()
        Generate graphs that visualise the search results.

    get_plot_objects():
        Get the plot objects for plots generated.

    _format_geo_query_string()
        formats the GEO DataSets portion of the input user query into a
        string.

    _format_geo_request()
        formats the GEO DataSets request payload

    _format_result(content)
        formats the search query output and converts it into a pandas
        dataframe

    See Also
    --------
    GeoSearch.info: GeoSearch usage details
    SraSearch: Superclass of GeoSearch
    QuerySearch: Superclass of SraSearch

    """

    def __init__(
        self,
        verbosity=2,
        return_max=20,
        query=None,
        accession=None,
        organism=None,
        layout=None,
        mbases=None,
        publication_date=None,
        platform=None,
        selection=None,
        source=None,
        strategy=None,
        title=None,
        geo_query=None,
        geo_dataset_type=None,
        geo_entry_type=None,
        suppress_validation=False,
    ):
        self.geo_fields = {
            "query": geo_query,
            "dataset_type": geo_dataset_type,
            "entry_type": geo_entry_type,
            "publication_date": publication_date,
            "organism": organism,
        }
        for k in self.geo_fields:
            if type(self.geo_fields[k]) == list:
                self.geo_fields[k] = " ".join(self.geo_fields[k])
        self.search_sra = True
        self.search_geo = True
        self.entries = {}
        self.number_entries = 0
        self.stats = {
            "study": "-",
            "experiment": "-",
            "run": "-",
            "sample": "-",
            "Date range": "-",
            "Organisms": "-",
            "Library strategy": "-",
            "Library source": "-",
            "Library selection": "-",
            "Library layout": "-",
            "Platform": "-",
            "count_mean": "-",
            "count_median": "-",
            "count_stdev": "-",
        }
        try:
            super().__init__(
                verbosity,
                return_max,
                query,
                accession,
                organism,
                layout,
                mbases,
                publication_date,
                platform,
                selection,
                source,
                strategy,
                title,
                suppress_validation,
            )
        except MissingQueryException:
            self.search_sra = False
        if not any(self.geo_fields.values()):
            self.search_geo = False
        if not self.search_geo and not self.search_sra:
            raise MissingQueryException()
        # Narrowing down the total number of eligible uids
        if self.fields["query"]:
            self.fields["query"] += " AND sra gds[Filter]"
        elif self.search_sra:
            self.fields["query"] = "sra gds[Filter]"
        if self.geo_fields["query"]:
            self.geo_fields["query"] += " AND gds sra[Filter]"
        elif self.search_geo:
            self.geo_fields["query"] = "gds sra[Filter]"

    def _format_geo_query_string(self):
        term = ""
        if self.geo_fields["query"]:
            term += self.geo_fields["query"] + " AND "
        if self.geo_fields["organism"]:
            term += self.geo_fields["organism"] + "[Organism] AND "
        if self.geo_fields["publication_date"]:
            dates = []
            for date in self.fields["publication_date"].split(":"):
                dates.append("/".join(date.split("-")[::-1]))
            term += ":".join(dates) + "[PDAT] AND "
        if self.geo_fields["dataset_type"]:
            term += self.geo_fields["dataset_type"] + "[DataSet Type] AND "
        if self.geo_fields["entry_type"]:
            term += self.geo_fields["entry_type"] + "[Entry Type] AND "
        return term[:-5]  # Removing trailing " AND "

    def _format_geo_request(self):
        payload = {
            "db": "gds",
            "term": self._format_geo_query_string(),
            "retmode": "json",
            "retmax": self.return_max * 10,
            "usehistory": "y",
        }
        return payload

    def _format_request(self):
        if not self.search_geo:
            retmax = self.return_max
        else:
            retmax = self.return_max * 10
        payload = {
            "db": "sra",
            "term": self._format_query_string(),
            "retmode": "json",
            "retmax": retmax,
        }
        return payload

    def search(self):
        if not self.search_geo:
            super().search()
        else:
            # Step 1: retrieves the list of uids from GEO DataSets, and use
            # ELink to find corresponding uids in SRA
            geo_payload = self._format_geo_request()
            try:
                r = requests_3_retries().get(
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
                    params=geo_payload,
                    timeout=SEARCH_REQUEST_TIMEOUT,
                )
                r.raise_for_status()
                result = r.json()["esearchresult"]
                query_key = result["querykey"]
                web_env = result["webenv"]
                elink_payload = {
                    "dbfrom": "gds",
                    "db": "sra",
                    "retmode": "json",
                    "query_key": query_key,
                    "WebEnv": web_env,
                }
                r = requests_3_retries().get(
                    "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/elink.fcgi",
                    params=elink_payload,
                    timeout=SEARCH_REQUEST_TIMEOUT,
                )
                r.raise_for_status()
                try:
                    data = r.json()
                    uids_from_geo = data["linksets"][0]["linksetdbs"][0]["links"]
                except (JSONDecodeError, KeyError, IndexError):
                    uids_from_geo = []
                # Step 2: Retrieve list of uids from SRA and
                # Find the intersection of both lists of uids
                if self.search_sra:
                    r = requests_3_retries().get(
                        "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi",
                        params=self._format_request(),
                        timeout=SEARCH_REQUEST_TIMEOUT,
                    )
                    r.raise_for_status()
                    uids_from_sra = r.json()["esearchresult"]["idlist"]
                    uids = list(set(uids_from_sra).intersection(uids_from_geo))
                else:
                    uids = uids_from_geo
                # Ensure that only return_max number of uids are used
                uids = uids[: self.return_max]
                # Step 3: retrieves the detailed information for each uid
                # returned, in groups of SRA_SEARCH_GROUP_SIZE.
                if not uids:
                    print(
                        f"No results found for the following search query: \n "
                        f"SRA: {self.fields}\nGEO DataSets: {self.geo_fields}"
                    )
                    return  # If no queries found, return nothing
                pbar = tqdm(total=len(uids))
                for i in range(0, len(uids), SRA_SEARCH_GROUP_SIZE):
                    current_uids = ",".join(
                        uids[i : min(i + SRA_SEARCH_GROUP_SIZE, len(uids))]
                    )
                    pbar.update(min(SRA_SEARCH_GROUP_SIZE, len(uids) - i))
                    payload2 = {"db": "sra", "retmode": "xml", "id": current_uids}

                    r = requests_3_retries().get(
                        "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi",
                        params=payload2,
                        timeout=SEARCH_REQUEST_TIMEOUT,
                        stream=True,
                    )
                    r.raise_for_status()
                    r.raw.decode_content = True
                    self._format_response(r.raw)
                pbar.close()
                self._format_result()
            except requests.exceptions.Timeout:
                sys.exit(f"Connection to the server has timed out. Please retry.")
            except requests.exceptions.HTTPError:
                sys.exit(
                    f"HTTPError: This is likely caused by an invalid search query: "
                    f"\nURL queried: {r.url} \nUser query: {self.fields}"
                )

    @classmethod
    def info(cls):
        """Information on how to use GeoSearch.

        Displays information on how to query GEO DataSets / SRA via
        GeoSearch, including accepted inputs for geo_query,
        geo_dataset_type and geo_entry_type.

        Returns
        -------
        info: str
            Information on how to use GeoSearch.
        """
        info = (
            "General Information:\n"
            "--------------------\n"
            "GeoSearch (Or 'pysradb search --db geo ...' on the command line) \n"
            "is able to query both SRA and GEO DataSets, returning the subset \n"
            "of entries that appears within both search queries. \n\n"
            "Queries sent to SRA and GEO DataSets will have 'sra gds[Filter]' \n"
            "and 'gds sra[Filter]' appended to the search queries respectively \n"
            "to ensure that the entries in the search result can also be found \n"
            "in the other API. \n\n"
            "Queries sent to SRA uses the same fields as SraSearch, \n"
            "or 'pysradb search --db sra ...' on the command line. \n\n"
            "The following fields, if used, are sent as part of the GEO DataSets query: \n"
            "organism, publication_date, geo_query, geo_dataset_type, geo_entry_type\n\n"
            "Notes about GEO DataSets specific fields: \n"
            "----------------------------------------- \n"
            "geo_query: This is the free text query, similar to 'query', that \n"
            "is sent to GEO DataSets API. The 'query' field is the free text \n"
            "query sent to SRA API instead.\n\n"
            "geo_dataset_type: The type of GEO DataSet, which can be one of the following: \n"
            "  expression profiling by array \n"
            "  expression profiling by genome tiling array \n"
            "  expression profiling by high throughput sequencing \n"
            "  expression profiling by mpss \n"
            "  expression profiling by rt pcr \n"
            "  expression profiling by sage \n"
            "  expression profiling by snp array \n"
            "  genome binding/occupancy profiling by array \n"
            "  genome binding/occupancy profiling by genome tiling array \n"
            "  genome binding/occupancy profiling by high throughput sequencing \n"
            "  genome binding/occupancy profiling by snp array \n"
            "  genome variation profiling by array \n"
            "  genome variation profiling by genome tiling array \n"
            "  genome variation profiling by high throughput sequencing \n"
            "  genome variation profiling by snp array \n"
            "  methylation profiling by array \n"
            "  methylation profiling by genome tiling array \n"
            "  methylation profiling by high throughput sequencing \n"
            "  methylation profiling by snp array \n"
            "  non coding rna profiling by array \n"
            "  non coding rna profiling by genome tiling array \n"
            "  non coding rna profiling by high throughput sequencing \n"
            "  other \n"
            "  protein profiling by mass spec \n"
            "  protein profiling by protein array \n"
            "  snp genotyping by snp array \n"
            "  third party reanalysis\n\n"
            "geo_dataset_type: The type of GEO entry, which can be one of the following: \n"
            "  gds\n"
            "  gpl\n"
            "  gse\n"
            "  gsm\n\n"
        )

        return info



import errno
import gzip
import io
import ntpath
import os
import shlex
import subprocess
import urllib.request as urllib_request
import warnings

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from tqdm.autonotebook import tqdm

from .exceptions import IncorrectFieldException

warnings.simplefilter(action="ignore", category=FutureWarning)


tqdm.pandas()


def path_leaf(path):
    """Get path's tail from a filepath.

    Parameters
    ----------
    path: string
          Filepath

    Returns
    -------
    tail: string
          Filename
    """
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)


def requests_3_retries():
    """Generates a requests session object that allows 3 retries.

    Returns
    -------
    session: requests.Session
        requests session object that allows 3 retries for server-side
        errors.
    """
    session = requests.Session()
    retry = Retry(
        total=3,
        backoff_factor=0.5,
        status_forcelist=[500, 502, 503, 504],
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


def scientific_name_to_taxid(name):
    """Converts a scientific name to its corresponding taxonomy ID.

    Parameters
    ----------
    name: str
        Scientific name of interest.

    Returns
    -------
    taxid: str
        Taxonomy Id of the Scientific name.

    Raises
    ------
    IncorrectFieldException
        If the scientific name cannot be found.

    """

    r = requests.get(
        "https://www.ebi.ac.uk/ena/data/taxonomy/v1/taxon/scientific-name/" + name,
        timeout=5,
    )
    if r.status_code == 404:
        raise IncorrectFieldException(f"Unknown scientific name: {name}")
    r.raise_for_status()
    return r.json()[0]["taxId"]


def unique(sequence):
    """Get unique elements from a list maintaining the order.

    Parameters
    ----------
    input_list: list

    Returns
    -------
    unique_list: list
                 List with unique elements maintaining the order
    """
    visited = set()
    return [x for x in sequence if not (x in visited or visited.add(x))]


class TqdmUpTo(tqdm):
    """Alternative Class-based version of the above.
    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.
    Inspired by [twine#242](https://github.com/pypa/twine/pull/242),
    [here](https://github.com/pypa/twine/commit/42e55e06).

    Credits:
    https://github.com/tqdm/tqdm/blob/69326b718905816bb827e0e66c5508c9c04bc06c/examples/tqdm_wget.py
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def _extract_first_field(data):
    """Extract first field from a list of fields."""
    return list(next(iter(zip(*data))))


def _find_aspera_keypath(aspera_dir=None):
    """Locate aspera key.

    Parameters
    ----------
    aspera_dir: string
                Location to aspera directory (optional)

    Returns
    -------
    aspera_keypath: string
                    Location to aspera key
    """
    if aspera_dir is None:
        aspera_dir = os.path.join(os.path.expanduser("~"), ".aspera")
    aspera_keypath = os.path.join(
        aspera_dir, "connect", "etc", "asperaweb_id_dsa.openssh"
    )
    if os.path.isfile(aspera_keypath):
        return aspera_keypath


def mkdir_p(path):
    """Python version mkdir -p

    Parameters
    ----------
    path : string
           Path to directory to create
    """
    if path:
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise


def order_dataframe(df, columns):
    """Order a dataframe

    Order a dataframe by moving the `columns` in the front

    Parameters
    ----------
    df: Dataframe
        Dataframe
    columns: list
             List of columns that need to be put in front
    """
    remaining_columns = [w for w in df.columns if w not in columns]
    df = df[columns + remaining_columns]
    return df


def _get_url(url, download_to, show_progress=True):
    """Download anything at a given url.

    Parameters
    ----------
    url: string
         http/https/ftp url
    download_to: string
                 File location to write the downloaded file to
    show_progress: bool
                   Set to True by default to print progress bar
    """
    desc_file = "Downloading {}".format(url.split("/")[-1])
    mkdir_p(os.path.dirname(download_to))
    if show_progress:
        with TqdmUpTo(
            unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=desc_file
        ) as t:
            urllib_request.urlretrieve(
                url, download_to, reporthook=t.update_to, data=None
            )
    else:
        urllib_request.urlretrieve(url, download_to)


def run_command(command, verbose=False):
    """Run a shell command"""
    process = subprocess.Popen(
        shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )

    while True:
        output = process.stdout.readline().strip()
        output = output.decode("utf-8")
        if output == "" and process.poll() is not None:
            break
        if output:
            if verbose:
                print((str(output.strip())))
    rc = process.poll()
    return rc


def get_gzip_uncompressed_size(filepath):
    """Get uncompressed size of a .gz file

    Parameters
    ----------
    filepath: string
              Path to input file

    Returns
    -------
    filesize: int
              Uncompressed file size
    """
    with gzip.open(filepath, "rb") as file_obj:
        return file_obj.seek(0, io.SEEK_END)


def confirm(preceeding_text):
    """Confirm user input.

    Parameters
    ----------
    preceeding_text: str
                     Text to print

    Returns
    -------
    response: bool
    """
    print(os.linesep, flush=True)
    notification_str = "Please respond with 'y' or 'n'"
    while True:
        choice = input("{} [Y/n]: ".format(preceeding_text)).lower()
        if choice in ["yes", "y"] or not choice:
            return True
        if choice in ["no", "n"]:
            return False
        print(notification_str, flush=True)


def copyfileobj(fsrc, fdst, bufsize=16384, filesize=None, desc=""):
    """Copy file object with a progress bar.

    Parameters
    ----------
    fsrc: filehandle
          Input file handle
    fdst: filehandle
          Output file handle
    bufsize: int
             Length of output buffer
    filesize: int
              Input file file size
    desc: string
          Description for tqdm status
    """
    with tqdm(
        total=filesize,
        unit="B",
        unit_scale=True,
        miniters=1,
        unit_divisor=1024,
        desc=desc,
    ) as pbar:
        while True:
            buf = fsrc.read(bufsize)
            if not buf:
                break
            fdst.write(buf)
            pbar.update(len(buf))
